#!/usr/bin/env python
# family_member.py
#
"""
.. module:: family_member
.. role:: red

BitDust family_member() Automat

EVENTS:
    * :red:`all-suppliers-agree`
    * :red:`contacts-received`
    * :red:`dht-read-fail`
    * :red:`dht-value-exist`
    * :red:`dht-value-not-exist`
    * :red:`dht-write-fail`
    * :red:`dht-write-ok`
    * :red:`disconnect`
    * :red:`family-join`
    * :red:`family-leave`
    * :red:`family-refresh`
    * :red:`init`
    * :red:`instant`
    * :red:`one-supplier-not-agree`
    * :red:`shutdown`
    * :red:`timer-10sec`
"""

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import re

from twisted.internet.task import LoopingCall

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.main import settings

from bitdust.lib import serialization
from bitdust.lib import strng
from bitdust.lib import nameurl

from bitdust.contacts import contactsdb

from bitdust.dht import dht_relations

from bitdust.userid import my_id
from bitdust.userid import id_url

from bitdust.raid import eccmap

from bitdust.p2p import p2p_service
from bitdust.p2p import commands

#------------------------------------------------------------------------------

_CustomersFamilies = {}

_ValidRequests = [
    'family-refresh',
    'family-join',
    'family-leave',
]

#------------------------------------------------------------------------------

DHT_RECORD_REFRESH_INTERVAL = 5*60

#------------------------------------------------------------------------------


def families():
    global _CustomersFamilies
    return _CustomersFamilies


def create_family(customer_idurl):
    customer_idurl = id_url.field(customer_idurl)
    if customer_idurl in families():
        raise Exception('another FamilyMember for %s already exists' % customer_idurl)
    families()[customer_idurl] = FamilyMember(customer_idurl)
    return families()[customer_idurl]


def delete_family(customer_idurl):
    customer_idurl = id_url.field(customer_idurl)
    if customer_idurl not in families():
        raise Exception('instance of FamilyMember for %s is not exist' % customer_idurl)
    families().pop(customer_idurl)
    return True


def by_customer_idurl(customer_idurl):
    customer_idurl = id_url.field(customer_idurl)
    return families().get(customer_idurl, None)


#------------------------------------------------------------------------------


class FamilyMember(automat.Automat):

    """
    This class implements all the functionality of ``family_member()`` state machine.
    """

    timers = {
        'timer-10sec': (10.0, ['SUPPLIERS']),
    }

    def __init__(self, customer_idurl, debug_level=_DebugLevel, log_events=_Debug, log_transitions=_Debug, publish_events=False, **kwargs):
        """
        Builds `family_member()` state machine.
        """
        self.customer_idurl = id_url.field(customer_idurl)
        self.supplier_idurl = my_id.getIDURL().to_bin()
        super(FamilyMember, self).__init__(
            name='family_%s_member_%s' % (
                nameurl.GetName(self.customer_idurl),
                nameurl.GetName(self.supplier_idurl),
            ), state='AT_STARTUP', debug_level=debug_level, log_events=log_events, log_transitions=log_transitions, publish_events=publish_events, **kwargs
        )

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        """
        Method to catch the moment when `family_member()` state were changed.
        """
        if event != 'instant' and newstate in [
            'CONNECTED',
            'DISCONNECTED',
        ]:
            self.automat('instant')

    def state_not_changed(self, curstate, event, *args, **kwargs):
        """
        This method intended to catch the moment when some event was fired in the `family_member()`
        but automat state was not changed.
        """
        if event != 'instant' and curstate in [
            'CONNECTED',
            'DISCONNECTED',
        ]:
            self.automat('instant')

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python <http://bitdust.io/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'DISCONNECTED'
                self.doInit(*args, **kwargs)
        #---DISCONNECTED---
        elif self.state == 'DISCONNECTED':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'instant' and self.isAnyRequests(*args, **kwargs):
                self.state = 'DHT_READ'
                self.Attempts = 0
                self.doPull(*args, **kwargs)
                self.doDHTRead(*args, **kwargs)
            elif event == 'family-refresh' or event == 'family-join' or event == 'family-leave':
                self.doPush(event, *args, **kwargs)
            elif event == 'contacts-received':
                self.doCheckReply(*args, **kwargs)
        #---DHT_READ---
        elif self.state == 'DHT_READ':
            if event == 'family-refresh' or event == 'family-join' or event == 'family-leave':
                self.doPush(event, *args, **kwargs)
            elif event == 'contacts-received':
                self.doCheckReply(*args, **kwargs)
            elif event == 'dht-read-fail':
                self.state = 'DISCONNECTED'
                self.Attempts = 0
                self.doNotifyDisconnected(*args, **kwargs)
            elif event == 'dht-value-exist' and self.isMyPositionOK(*args, **kwargs) and not self.isLeaving(*args, **kwargs):
                self.state = 'CONNECTED'
                self.Attempts = 0
                self.doNotifyConnected(*args, **kwargs)
            elif event == 'dht-value-not-exist' or (event == 'dht-value-exist' and not self.isMyPositionOK(*args, **kwargs)):
                self.state = 'SUPPLIERS'
                self.Attempts += 1
                self.doRebuildFamily(*args, **kwargs)
                self.doRequestSuppliersReview(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'dht-value-exist' and self.isMyPositionOK(*args, **kwargs) and self.isLeaving(*args, **kwargs):
                self.state = 'DHT_WRITE'
                self.doRebuildFamily(*args, **kwargs)
                self.doDHTWrite(*args, **kwargs)
        #---SUPPLIERS---
        elif self.state == 'SUPPLIERS':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'family-refresh' or event == 'family-join' or event == 'family-leave':
                self.doPush(event, *args, **kwargs)
            elif event == 'contacts-received':
                self.doCheckReply(*args, **kwargs)
            elif (event == 'all-suppliers-agree' or event == 'timer-10sec') and not self.isFamilyModified(*args, **kwargs):
                self.state = 'CONNECTED'
                self.Attempts = 0
                self.doNotifyConnected(*args, **kwargs)
            elif (event == 'timer-10sec' or event == 'all-suppliers-agree') and self.isFamilyModified(*args, **kwargs):
                self.state = 'DHT_WRITE'
                self.doDHTWrite(*args, **kwargs)
            elif event == 'one-supplier-not-agree':
                self.doSolveConflict(*args, **kwargs)
                self.doRequestSuppliersReview(*args, **kwargs)
        #---DHT_WRITE---
        elif self.state == 'DHT_WRITE':
            if event == 'family-refresh' or event == 'family-join' or event == 'family-leave':
                self.doPush(event, *args, **kwargs)
            elif event == 'contacts-received':
                self.doCheckReply(*args, **kwargs)
            elif event == 'dht-write-fail' and not self.isLeaving(*args, **kwargs) and self.Attempts > 3:
                self.state = 'DISCONNECTED'
                self.Attempts = 0
                self.doNotifyDisconnected(*args, **kwargs)
            elif event == 'dht-write-fail' and not self.isLeaving(*args, **kwargs) and self.Attempts <= 3:
                self.state = 'DHT_READ'
                self.doDHTRead(*args, **kwargs)
            elif event == 'shutdown' or ((event == 'dht-write-fail' or event == 'dht-write-ok') and self.isLeaving(*args, **kwargs)):
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'dht-write-ok' and not self.isLeaving(*args, **kwargs):
                self.state = 'CONNECTED'
                self.Attempts = 0
                self.doNotifyConnected(*args, **kwargs)
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        #---CONNECTED---
        elif self.state == 'CONNECTED':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'disconnect':
                self.state = 'DISCONNECTED'
                self.doNotifyDisconnected(*args, **kwargs)
            elif event == 'instant' and self.isAnyRequests(*args, **kwargs):
                self.state = 'DHT_READ'
                self.Attempts = 0
                self.doPull(*args, **kwargs)
                self.doDHTRead(*args, **kwargs)
            elif event == 'family-refresh' or event == 'family-join' or event == 'family-leave':
                self.doPush(event, *args, **kwargs)
            elif event == 'contacts-received':
                self.doCheckReply(*args, **kwargs)
        return None

    def isAnyRequests(self, *args, **kwargs):
        """
        Condition method.
        """
        return len(self.requests) > 0

    def isLeaving(self, *args, **kwargs):
        """
        Condition method.
        """
        return self.current_request and self.current_request['command'] == 'family-leave'

    def isMyPositionOK(self, *args, **kwargs):
        """
        Condition method.
        """
        dht_info_valid = self._do_validate_dht_info(args[0])
        if not dht_info_valid:
            return False
        if self.current_request and self.current_request['command'] == 'family-leave':
            if my_id.getIDURL().to_bin() not in dht_info_valid['suppliers']:
                return True
        my_info_valid = self._do_validate_my_info(self.my_info)
        if not my_info_valid:
            return False
        latest_revision = self._do_detect_latest_revision(dht_info_valid, my_info_valid)
        if latest_revision == 0:
            return False
        try:
            my_position = my_info_valid['suppliers'].index(my_id.getIDURL().to_bin())
        except:
            my_position = -1
        if my_position < 0:
            return False
        try:
            existing_position = dht_info_valid['suppliers'].index(my_id.getIDURL().to_bin())
        except:
            existing_position = -1
        return existing_position > 0 and my_position > 0 and existing_position == my_position

    def isFamilyModified(self, *args, **kwargs):
        """
        Condition method.
        """
        return self.transaction is not None

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        self.requests = []
        self.current_request = None
        self.dht_info = None
        self.my_info = None
        self.transaction = None
        self.dht_value_exists = False
        self.dht_read_use_cache = True
        self.refresh_period = DHT_RECORD_REFRESH_INTERVAL*settings.DefaultDesiredSuppliers()
        self.refresh_task = LoopingCall(self._on_family_refresh_task)

    def doPush(self, event, *args, **kwargs):
        """
        Action method.
        """
        if event not in _ValidRequests:
            raise Exception('Invalid request: %r' % args)
        request = (args[0] if args else {}) or {}
        request['command'] = event
        self.requests.append(request)

    def doPull(self, *args, **kwargs):
        """
        Action method.
        """
        self.current_request = self.requests.pop(0)

    def doRebuildFamily(self, *args, **kwargs):
        """
        Action method.
        """
        dht_info_valid = self._do_validate_dht_info(args[0])
        my_info_valid = self._do_validate_my_info(self.my_info)
        latest_revision = self._do_detect_latest_revision(dht_info_valid, my_info_valid)
        merged_info = None
        if latest_revision > 0:
            merged_info = self._do_merge_revisions(dht_info_valid, my_info_valid, latest_revision)
        if not merged_info:
            merged_info = self._do_create_first_revision(self.current_request)


#         if not merged_info:
#             lg.err('failed to merge customer family info after reading from DHT, skip transaction')
#             self.transaction = None
#             return
        possible_transaction = self._do_process_request(merged_info, self.current_request)
        if not possible_transaction:
            lg.warn('failed to process customer family %r change request, skip transaction' % self)
            return
        self.transaction = self._do_increment_revision(possible_transaction)
        if _Debug:
            lg.args(_DebugLevel, transaction=self.transaction)
        if self.transaction:
            known_ecc_map = self.transaction.get('ecc_map')
            if known_ecc_map:
                expected_suppliers_count = eccmap.GetEccMapSuppliersNumber(known_ecc_map)
                self.refresh_period = DHT_RECORD_REFRESH_INTERVAL*expected_suppliers_count

    def doRequestSuppliersReview(self, *args, **kwargs):
        """
        Action method.
        """
        if not self.transaction:
            self.automat('all-suppliers-agree')
            return
        self.suppliers_requests = []
        for supplier_idurl in self.transaction['suppliers']:
            if not supplier_idurl:
                continue
            if supplier_idurl == my_id.getIDURL().to_bin():
                continue
            outpacket = p2p_service.SendContacts(
                remote_idurl=supplier_idurl,
                json_payload={
                    'space': 'family_member',
                    'type': 'suppliers_list',
                    'customer_idurl': self.customer_idurl,
                    'customer_ecc_map': self.transaction['ecc_map'],
                    'transaction_revision': self.transaction['revision'],
                    'suppliers_list': self.transaction['suppliers'],
                },
                callbacks={
                    commands.Ack(): self._on_supplier_ack,
                    commands.Fail(): self._on_supplier_fail,
                },
            )
            self.suppliers_requests.append(outpacket.PacketID)
        if not self.suppliers_requests:
            self.automat('all-suppliers-agree')
        else:
            if _Debug:
                lg.out(_DebugLevel, 'family_member.doRequestSuppliersReview sent to transaction for review to %d suppliers' % len(self.suppliers_requests))

    def doSolveConflict(self, *args, **kwargs):
        """
        Action method.
        """
        # TODO: take in account ecc_map while solving the conflict
        # ecc_map = kwargs.get('ecc_map')
        suppliers_list = kwargs.get('suppliers_list')
        another_supplier_idurl = kwargs.get('supplier_idurl')
        try:
            another_supplier_position = suppliers_list.index(another_supplier_idurl)
        except:
            another_supplier_position = -1
        if another_supplier_position < 0:
            # this must never happen actually... only if another supplier is really uncooperative
            # this is dangerous because can lead to infinite loop between me and another supplier
            lg.err('found uncooperative supplier %s who raised the conflict but replied with invalid response' % another_supplier_idurl)
            # TODO: solve later
            self.transaction = None
        if self.transaction:
            if len(self.transaction['suppliers']) <= another_supplier_position:
                lg.warn('another supplier position larger than family size, failed to solve family conflict with supplier %s' % another_supplier_idurl)
                self.transaction = None
            else:
                if self.transaction['suppliers'][another_supplier_position]:
                    lg.warn('given position is not empty, failed to solve family conflict with supplier %s' % another_supplier_idurl)
                    self.transaction = None
                else:
                    self.transaction['suppliers'][another_supplier_position] = another_supplier_idurl
                    lg.info('found desired position %d in the family and solved conflict with supplier %s' % (another_supplier_position, another_supplier_idurl))

    def doDHTRead(self, *args, **kwargs):
        """
        Action method.
        """
        d = dht_relations.read_customer_suppliers(self.customer_idurl, as_fields=False, use_cache=False)
        d.addCallback(self._on_dht_read_success)
        d.addErrback(self._on_dht_read_failed)

    def doDHTWrite(self, *args, **kwargs):
        """
        Action method.
        """
        self._do_write_transaction(0)

    def doNotifyConnected(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'family_member.doNotifyConnected\n            my_info=%r\n            dht_info=%r\n            requests=%r' % (self.my_info, self.dht_info, self.requests))
        to_be_closed = False
        if self.current_request['command'] == 'family-leave':
            to_be_closed = True
        self.current_request = None
        if self.refresh_task.running:
            self.refresh_task.stop()
        self.refresh_task.start(self.refresh_period, now=False)
        if to_be_closed:
            self.requests = []
            self.automat('shutdown')

    def doNotifyDisconnected(self, *args, **kwargs):
        """
        Action method.
        """
        if self.refresh_task.running:
            self.refresh_task.stop()
        self.current_request = None

    def doCheckReply(self, *args, **kwargs):
        """
        Action method.
        """
        self._on_incoming_contacts_packet(args[0])

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        self.requests = []
        self.current_request = None
        self.my_info = None
        self.dht_info = None
        self.transaction = None
        if self.refresh_task.running:
            self.refresh_task.stop()
        self.refresh_task = None
        delete_family(self.customer_idurl)
        self.destroy()

    #------------------------------------------------------------------------------

    def _do_validate_dht_info(self, inp):
        if _Debug:
            lg.args(_DebugLevel, inp=inp)
        if not inp or not isinstance(inp, dict):
            self.dht_read_use_cache = False
            return None
        out = inp.copy()
        try:
            dht_revision = int(out['revision'])
            out['suppliers'] = id_url.to_bin_list(out['suppliers'])
            ecc_map = out['ecc_map']
            if dht_revision < 1:
                raise Exception('invalid revision')
            if not isinstance(out['suppliers'], list) or len(out['suppliers']) < 1:
                raise Exception('family %r must include some suppliers' % self)
            if ecc_map and ecc_map not in eccmap.EccMapNames():
                raise Exception('invalid ecc_map name in family %r' % self)
            out['publisher_idurl'] = id_url.field(out['publisher_idurl'])
            # TODO: add publisher_signature and Validate method to check publisher signature
            out['customer_idurl'] = id_url.field(out['customer_idurl'])
            # TODO: add customer_signature and Validate method to check customer signature
        except:
            lg.exc()
            lg.warn('skip invalid DHT info and assume DHT record is not exist')
            self.dht_read_use_cache = False
            return None
        return out

    def _do_validate_my_info(self, inp):
        if not inp:
            return None
        if not inp or not isinstance(inp, dict):
            return None
        out = inp.copy()
        try:
            my_revision = int(out['revision'])
            if my_revision < 1:
                raise Exception('invalid revision')
        except Exception as exc:
            lg.warn(str(exc))
            return None
        try:
            out['suppliers'] = id_url.to_bin_list(out['suppliers'])
            if not isinstance(out['suppliers'], list) or len(out['suppliers']) < 1:
                raise Exception('family %r must include some suppliers' % self)
        except Exception as exc:
            lg.warn(str(exc))
            return None
        try:
            ecc_map = out['ecc_map']
            if ecc_map and ecc_map not in eccmap.EccMapNames():
                raise Exception('invalid ecc_map name in family %r' % self)
        except Exception as exc:
            lg.warn(str(exc))
            return None
        try:
            out['publisher_idurl'] = id_url.field(out['publisher_idurl'])
            # TODO: if I am a publisher - revision number must be the same as my info
        except Exception as exc:
            lg.warn(str(exc))
            return None
        try:
            customer_idurl = id_url.field(out['customer_idurl'])
            if customer_idurl != self.customer_idurl:
                raise Exception('invalid customer_idurl')
        except Exception as exc:
            lg.warn(str(exc))
            return None
        return out

    def _do_create_first_revision(self, request):
        merged_info = {
            'revision': 0,
            'publisher_idurl': my_id.getIDURL(),  # I will be a publisher of the first revision
            'suppliers': id_url.to_bin_list(request.get('family_snapshot') or []),
            'ecc_map': request.get('ecc_map'),
            'customer_idurl': self.customer_idurl,
        }
        if _Debug:
            lg.args(_DebugLevel, merged_info=merged_info)
        return merged_info

    def _do_create_possible_revision(self, latest_revision):
        local_customer_meta_info = contactsdb.get_customer_meta_info(self.customer_idurl)
        try:
            possible_position = int(local_customer_meta_info.get('position', -1))
        except:
            possible_position = -1
        possible_suppliers = id_url.to_bin_list(local_customer_meta_info.get('family_snapshot') or [])
        if possible_position >= 0 and my_id.getIDURL().to_bin() not in possible_suppliers:
            if len(possible_suppliers) > possible_position:
                possible_suppliers[possible_position] = my_id.getIDURL().to_bin()
        my_info = {
            'revision': latest_revision,
            'publisher_idurl': my_id.getIDURL(),  # I will be a publisher of that revision
            'suppliers': possible_suppliers,
            'ecc_map': local_customer_meta_info.get('ecc_map'),
            'customer_idurl': self.customer_idurl,
        }
        if _Debug:
            lg.args(_DebugLevel, my_info=my_info)
        return my_info

    def _do_create_revision_from_another_supplier(self, another_revision, another_suppliers, another_ecc_map):
        local_customer_meta_info = contactsdb.get_customer_meta_info(self.customer_idurl)
        try:
            possible_position = int(local_customer_meta_info.get('position', -1))
        except:
            possible_position = -1
        if possible_position >= 0:
            try:
                another_suppliers[possible_position] = my_id.getIDURL().to_bin()
            except:
                lg.exc()
            contactsdb.add_customer_meta_info(
                self.customer_idurl,
                {
                    'ecc_map': another_ecc_map,
                    'position': possible_position,
                    'family_snapshot': id_url.to_bin_list(another_suppliers),
                },
            )
        return {
            'revision': int(another_revision),
            'publisher_idurl': my_id.getIDURL(),  # I will be a publisher of that revision
            'suppliers': id_url.to_bin_list(another_suppliers),
            'ecc_map': another_ecc_map,
            'customer_idurl': self.customer_idurl,
        }

    def _do_detect_latest_revision(self, dht_info, my_info):
        try:
            my_revision = int(my_info['revision'])
        except:
            lg.warn('my own info is unknown or invalid, assume my revision is 0')
            my_revision = 0
        try:
            dht_revision = int(dht_info['revision'])
        except:
            lg.warn('DHT info is unknown or invalid, assume DHT revision is 0')
            dht_revision = 0
        if my_revision == dht_revision:
            return dht_revision
        if my_revision > dht_revision:
            # TODO: SECURITY need to find a solution to prevent cheating here
            # another supplier could publish a record where he is only alone present and with a correct revision
            # that means he actually brutally dropped all other suppliers from the family
            lg.info('known DHT info for customer %s is more fresh, will rewrite DHT record' % self.customer_idurl)
            if my_revision > dht_revision + 1:
                lg.warn('switching revision too far, normally always increase by one on every change')
            return my_revision
        return dht_revision

    def _do_merge_revisions(self, dht_info, my_info, latest_revision):
        if _Debug:
            lg.args(_DebugLevel, dht_info=dht_info, my_info=my_info, latest_revision=latest_revision)
        if dht_info is None or not isinstance(dht_info, dict):
            merged_info = my_info
        else:
            if latest_revision == int(dht_info['revision']):
                if my_info is not None:
                    if latest_revision == int(my_info['revision']):
                        # I have same revision as info from DHT
                        merged_info = dht_info
                    else:
                        if int(my_info['revision']) > int(dht_info['revision']):
                            # here my revision is higher, so I have some changes that needs to be published already
                            merged_info = my_info
                        else:
                            # here my revision is lower, I need to take info from DHT
                            merged_info = dht_info
                else:
                    merged_info = dht_info
            else:
                # here my revision is higher, so I have some changes that needs to be published already
                merged_info = my_info
        if not merged_info:
            return None
        # make sure list of suppliers have correct length according to ecc_map
        if not merged_info['ecc_map']:
            known_ecc_map = contactsdb.get_customer_meta_info(self.customer_idurl).get('ecc_map', None)
            lg.warn('unknown ecc_map, will populate known value: %s' % known_ecc_map)
            merged_info['ecc_map'] = known_ecc_map
        if merged_info['ecc_map']:
            expected_suppliers_count = eccmap.GetEccMapSuppliersNumber(merged_info['ecc_map'])
            if len(merged_info['suppliers']) < expected_suppliers_count:
                merged_info['suppliers'] += [
                    b'',
                ]*(expected_suppliers_count - len(merged_info['suppliers']))
            elif len(merged_info['suppliers']) > expected_suppliers_count:
                merged_info['suppliers'] = merged_info['suppliers'][:expected_suppliers_count]
        if merged_info['revision'] != latest_revision:
            lg.info('will switch known revision %d to the latest: %d' % (merged_info['revision'], latest_revision))
        merged_info['revision'] = latest_revision
        if _Debug:
            lg.out(_DebugLevel, '    merged_info=%r' % merged_info)
        return merged_info

    def _do_increment_revision(self, possible_transaction):
        if self.customer_idurl.to_original() != possible_transaction['customer_idurl'].to_original():
            possible_transaction['customer_idurl'] = self.customer_idurl
            possible_transaction['revision'] += 1
            possible_transaction['publisher_idurl'] = my_id.getIDURL()
            lg.info('incremented family revision after customer %r identity rotated: %r' % (self.customer_idurl, possible_transaction['revision']))
            return possible_transaction
        if self.dht_info:
            if self.dht_info['suppliers'] == possible_transaction['suppliers']:
                if self.dht_info['ecc_map'] == possible_transaction['ecc_map']:
                    if self.current_request and self.current_request['command'] == 'family-leave':
                        if _Debug:
                            lg.out(_DebugLevel, 'family_member._do_increment_revision will re-publish latest DHT info because processing "family-leave" request')
                    else:
                        if _Debug:
                            lg.out(_DebugLevel, 'family_member._do_increment_revision did not found any changes, skip transaction')
                        return None
        possible_transaction['revision'] += 1
        possible_transaction['publisher_idurl'] = my_id.getIDURL()
        return possible_transaction

    def _do_process_family_join_request(self, merged_info, current_request):
        if _Debug:
            lg.args(_DebugLevel, merged_info=merged_info, current_request=current_request)
        current_request_expected_suppliers_count = None
        if current_request['ecc_map']:
            current_request_expected_suppliers_count = eccmap.GetEccMapSuppliersNumber(current_request['ecc_map'])
        if current_request_expected_suppliers_count and current_request.get('position') and current_request['position'] >= current_request_expected_suppliers_count:
            lg.warn('"family-join" request is not valid, supplier position %d greater than expected suppliers count %d for %s' % (current_request['position'], current_request_expected_suppliers_count, current_request['ecc_map']))
            return None

        if merged_info['ecc_map'] and current_request['ecc_map'] and current_request['ecc_map'] != merged_info['ecc_map']:
            lg.info('from "family-join" request, detected ecc_map change %s -> %s for customer %s' % (merged_info['ecc_map'], current_request['ecc_map'], self.customer_idurl))
            merged_info['ecc_map'] = current_request['ecc_map']
        if not merged_info['ecc_map'] and current_request['ecc_map']:
            lg.info('from "family-join" request, detected ecc_map was set to %s for the first time for customer %s' % (current_request['ecc_map'], self.customer_idurl))
            merged_info['ecc_map'] = current_request['ecc_map']
        if not merged_info['ecc_map']:
            known_ecc_map = contactsdb.get_customer_meta_info(self.customer_idurl).get('ecc_map', None)
            if known_ecc_map:
                lg.warn('unknown ecc_map, will populate known value from customer meta info: %s' % known_ecc_map)
                merged_info['ecc_map'] = known_ecc_map
        if not merged_info['ecc_map']:
            lg.warn('still did not found actual ecc_map from DHT or from the request')
            return None

        expected_suppliers_count = eccmap.GetEccMapSuppliersNumber(merged_info['ecc_map'])
        if not merged_info['suppliers']:
            merged_info['suppliers'] = [
                b'',
            ]*expected_suppliers_count

        if len(merged_info['suppliers']) < expected_suppliers_count:
            merged_info['suppliers'] += [
                b'',
            ]*(expected_suppliers_count - len(merged_info['suppliers']))
        else:
            merged_info['suppliers'] = merged_info['suppliers'][:expected_suppliers_count]

        try:
            existing_position = merged_info['suppliers'].index(current_request['supplier_idurl'])
        except:
            existing_position = -1

        if current_request['position'] is not None and current_request['position'] >= 0:
            if current_request['position'] >= expected_suppliers_count:
                lg.warn('"family-join" request is not valid, supplier position greater than expected suppliers count')
                return None
            if existing_position >= 0 and existing_position != current_request['position']:
                merged_info['suppliers'][existing_position] = b''
                merged_info['suppliers'][current_request['position']] = current_request['supplier_idurl']
                if _Debug:
                    lg.out(_DebugLevel, '    found my IDURL on %d position and will move it on %d position in the family of customer %s' % (existing_position, current_request['position'], self.customer_idurl))
            if merged_info['suppliers'][current_request['position']] != current_request['supplier_idurl']:
                if merged_info['suppliers'][current_request['position']]:
                    # TODO: SECURITY need to implement a signature verification and
                    # also build solution to validate that change was approved by customer
                    lg.warn('overwriting another supplier %s with my IDURL at position %d in family of customer %s' % (merged_info['suppliers'][current_request['position']], current_request['position'], self.customer_idurl))
                merged_info['suppliers'][current_request['position']] = current_request['supplier_idurl']
                if _Debug:
                    lg.out(_DebugLevel, '    placed supplier %s at known position %d in the family of customer %s' % (current_request['supplier_idurl'], current_request['position'], self.customer_idurl))

        if current_request['supplier_idurl'] not in merged_info['suppliers']:
            if b'' in merged_info['suppliers']:
                first_empty_position = merged_info['suppliers'].index(b'')
                merged_info['suppliers'][first_empty_position] = current_request['supplier_idurl']
                if _Debug:
                    lg.out(_DebugLevel, '    placed supplier %s at first empty position %d in family of customer %s' % (current_request['supplier_idurl'], first_empty_position, self.customer_idurl))
            else:
                merged_info['suppliers'].append(current_request['supplier_idurl'])
                if _Debug:
                    lg.out(_DebugLevel, '    added supplier %s to family of customer %s' % (current_request['supplier_idurl'], self.customer_idurl))

        if current_request.get('family_snapshot'):
            for supplier_position in range(len(merged_info['suppliers'])):
                if supplier_position < len(current_request['family_snapshot']):
                    if not merged_info['suppliers'][supplier_position] and current_request['family_snapshot'][supplier_position]:
                        merged_info['suppliers'][supplier_position] = current_request['family_snapshot'][supplier_position]
                        if _Debug:
                            lg.out(_DebugLevel, '    found empty supplier at position %d and populated from current request: %s' % (supplier_position, merged_info['suppliers'][supplier_position]))

        return merged_info

    def _do_process_family_leave_request(self, merged_info, current_request):
        if _Debug:
            lg.args(_DebugLevel, merged_info=merged_info, current_request=current_request)
        try:
            existing_position = merged_info['suppliers'].index(current_request['supplier_idurl'])
        except ValueError:
            existing_position = -1

        if current_request.get('ecc_map'):
            if merged_info['ecc_map'] and current_request.get('ecc_map') and current_request.get('ecc_map') != merged_info['ecc_map']:
                lg.info('from "family-leave" request, detected ecc_map change %s -> %s for customer %s' % (merged_info['ecc_map'], current_request['ecc_map'], self.customer_idurl))
                merged_info['ecc_map'] = current_request['ecc_map']
            if not merged_info['ecc_map'] and current_request['ecc_map']:
                lg.info('from "family-leave" request, detected ecc_map was set to %s for the first time for customer %s' % (current_request['ecc_map'], self.customer_idurl))
                merged_info['ecc_map'] = current_request['ecc_map']

        if not merged_info['ecc_map']:
            lg.warn('still did not found actual ecc_map from DHT or from the request')
            return None

        expected_suppliers_count = eccmap.GetEccMapSuppliersNumber(merged_info['ecc_map'])
        if not merged_info['suppliers']:
            merged_info['suppliers'] = [
                b'',
            ]*expected_suppliers_count

        if len(merged_info['suppliers']) < expected_suppliers_count:
            merged_info['suppliers'] += [
                b'',
            ]*(expected_suppliers_count - len(merged_info['suppliers']))
        else:
            merged_info['suppliers'] = merged_info['suppliers'][:expected_suppliers_count]

        if existing_position < 0:
            if _Debug:
                lg.dbg(_DebugLevel, 'supplier %r not found in customer family %r, probably already left' % (current_request['supplier_idurl'], self.customer_idurl))
        else:
            if existing_position < expected_suppliers_count:
                merged_info['suppliers'][existing_position] = b''
                if _Debug:
                    lg.info('erasing supplier %r from customer family %r' % (current_request['supplier_idurl'], self.customer_idurl))
        return merged_info

    def _do_process_family_refresh_request(self, merged_info):
        if _Debug:
            lg.args(_DebugLevel, merged_info=merged_info, my_info=self.my_info)
        if not self.my_info:
            self.my_info = self._do_create_possible_revision(int(merged_info['revision']))
            lg.warn('"family-refresh" request will use "possible" customer meta info: %r' % self.my_info)

        if int(self.my_info['revision']) > int(merged_info['revision']):
            lg.info('"family-refresh" request will overwrite DHT record with my info because my revision is higher than record in DHT')
            return self.my_info.copy()

        try:
            my_position = self.my_info['suppliers'].index(my_id.getIDURL().to_bin())
        except:
            my_position = -1
        if my_position < 0:
            lg.warn('"family-refresh" request failed because my info not exist or not valid, my own position in the family is unknown')
            return None

        my_expected_suppliers_count = None
        if self.my_info['ecc_map']:
            my_expected_suppliers_count = eccmap.GetEccMapSuppliersNumber(self.my_info['ecc_map'])
        if my_expected_suppliers_count and my_position >= my_expected_suppliers_count:
            lg.warn('"family-refresh" request failed because my info is not valid, supplier position greater than expected suppliers count')
            return None

        if my_expected_suppliers_count and len(merged_info['suppliers']) != my_expected_suppliers_count:
            lg.warn('number of suppliers not expected during processing of "family-refresh" request')
            if len(merged_info['suppliers']) < my_expected_suppliers_count:
                merged_info['suppliers'] += [
                    b'',
                ]*(my_expected_suppliers_count - len(merged_info['suppliers']))
            else:
                merged_info['suppliers'] = merged_info['suppliers'][:my_expected_suppliers_count]

        try:
            existing_position = merged_info['suppliers'].index(my_id.getIDURL().to_bin())
        except ValueError:
            existing_position = -1
        if existing_position < 0:
            if merged_info['suppliers'][my_position]:
                # TODO: SECURITY need to implement a signature verification and
                # also build solution to validate that change was approved by customer
                lg.warn('overwriting another supplier %s with my IDURL at position %d in family of customer %s' % (merged_info['suppliers'][my_position], my_position, self.customer_idurl))
            merged_info['suppliers'][my_position] = my_id.getIDURL().to_bin()
            if _Debug:
                lg.out(_DebugLevel, '    placed supplier %s at known position %d in the family of customer %s' % (my_id.getIDURL(), my_position, self.customer_idurl))
            existing_position = my_position

        if existing_position != my_position:
            merged_info['suppliers'][existing_position] = b''
            merged_info['suppliers'][my_position] = my_id.getIDURL().to_bin()
            if _Debug:
                lg.out(_DebugLevel, '    found my IDURL on %d position and will move it on %d position in the family of customer %s' % (existing_position, my_position, self.customer_idurl))
        return merged_info

    def _do_process_request(self, merged_info, current_request):
        if current_request['command'] == 'family-join':
            return self._do_process_family_join_request(merged_info, current_request)
        if current_request['command'] == 'family-leave':
            return self._do_process_family_leave_request(merged_info, current_request)
        if current_request['command'] == 'family-refresh':
            return self._do_process_family_refresh_request(merged_info)
        lg.err('invalid request command')
        return None

    def _do_write_transaction(self, retries):
        if _Debug:
            lg.out(_DebugLevel, 'family_member._do_write_transaction  suppliers=%d  retries=%d' % (len(self.transaction['suppliers']), retries))
        d = dht_relations.write_customer_suppliers(
            customer_idurl=self.customer_idurl,
            suppliers_list=self.transaction['suppliers'],
            ecc_map=self.transaction['ecc_map'],
            revision=self.transaction['revision'],
            publisher_idurl=self.transaction['publisher_idurl'],
        )
        d.addCallback(self._on_dht_write_success)
        d.addErrback(self._on_dht_write_failed, retries)

    def _on_family_refresh_task(self):
        if self.state == 'CONNECTED':
            self.automat('family-refresh')

    def _on_dht_read_success(self, dht_result):
        if dht_result and isinstance(dht_result, dict) and len(dht_result.get('suppliers', [])) > 0:
            if _Debug:
                lg.out(_DebugLevel, 'family_member._on_dht_read_success  result with %d suppliers' % len(dht_result.get('suppliers', [])))
            self.dht_info = dht_result
            self.dht_value_exists = True
            self.automat('dht-value-exist', dht_result)
        else:
            if _Debug:
                lg.out(_DebugLevel, 'family_member._on_dht_read_success  result with %s' % type(dht_result))
            self.dht_info = None
            self.dht_value_exists = False
            self.automat('dht-value-not-exist', None)

    def _on_dht_read_failed(self, err):
        self.dht_info = None
        self.dht_value_exists = False
        if _Debug:
            lg.out(_DebugLevel, 'family_member._on_dht_read_failed : %r' % err)
        self.automat('dht-read-fail')

    def _on_dht_write_success(self, dht_result):
        if _Debug:
            lg.out(_DebugLevel, 'family_member._on_dht_write_success  result: %r' % type(dht_result))
        self.my_info = self.transaction.copy()
        self.dht_info = None
        self.transaction = None
        self.dht_read_use_cache = False
        self.automat('dht-write-ok', dht_result)

    def _on_dht_write_failed(self, err, retries):
        if _Debug:
            lg.out(_DebugLevel, 'family_member._on_dht_write_failed  err: %r' % err)
        self.dht_read_use_cache = False
        try:
            errmsg = err.value.subFailure.getErrorMessage()
        except:
            try:
                errmsg = err.getErrorMessage()
            except:
                try:
                    errmsg = err.value
                except:
                    errmsg = str(err)
        err_msg = strng.to_text(errmsg)
        lg.err('failed to write family info for %s : %s' % (self.customer_idurl, err_msg))
        if err_msg.count('current revision is') and retries < 3:
            try:
                current_revision = re.search('current revision is (\d+)', err_msg).group(1)
                current_revision = int(current_revision)
            except:
                lg.exc()
                current_revision = self.transaction['revision']
            current_revision += 1
            self.transaction['revision'] = current_revision
            if _Debug:
                lg.warn('recognized "DHT write operation failed" because of late revision, increase revision to %d and retry' % current_revision)
            self._do_write_transaction(retries + 1)
            return
        self.transaction = None
        self.dht_info = None
        self.automat('dht-write-fail')

    def _on_incoming_suppliers_list(self, inp):
        incoming_packet = inp['packet']
        if _Debug:
            lg.out(_DebugLevel, 'family_member._on_incoming_suppliers_list with %s' % incoming_packet)
        if not self.my_info:
            if _Debug:
                lg.out(_DebugLevel, '    current DHT info is not yet known, skip')
            return p2p_service.SendAck(incoming_packet)
        try:
            another_ecc_map = inp['customer_ecc_map']
            another_suppliers_list = id_url.to_bin_list(inp['suppliers_list'])
            another_revision = int(inp['transaction_revision'])
        except:
            lg.exc()
            return p2p_service.SendFail(incoming_packet, response=serialization.DictToBytes(self.my_info))
        if _Debug:
            lg.args(_DebugLevel, another_revision=another_revision, another_ecc_map=another_ecc_map, another_suppliers_list=another_suppliers_list)
        if another_revision >= int(self.my_info['revision']):
            self.my_info = self._do_create_revision_from_another_supplier(another_revision, another_suppliers_list, another_ecc_map)
            lg.info('another supplier have more fresh revision, update my info and raise "family-refresh" event')
            self.automat('family-refresh')
            return p2p_service.SendAck(incoming_packet)
        if my_id.getIDURL().to_bin() not in another_suppliers_list:
            if _Debug:
                lg.args(_DebugLevel, another_suppliers_list=another_suppliers_list, customer_idurl=self.customer_idurl)
            lg.warn('another supplier is trying to remove my IDURL from the family of customer %s' % self.customer_idurl)
            return p2p_service.SendFail(incoming_packet, response=serialization.DictToBytes(self.my_info))
        my_position_in_transaction = another_suppliers_list.index(my_id.getIDURL().to_bin())
        try:
            my_known_position = self.my_info['suppliers'].index(my_id.getIDURL().to_bin())
        except:
            my_known_position = -1
        if my_known_position < 0:
            if _Debug:
                lg.args(_DebugLevel, my_known_position=my_known_position, another_suppliers_list=another_suppliers_list, customer_idurl=self.customer_idurl)
            lg.warn('another supplier is trying to remove me from the family of customer %s' % self.customer_idurl)
            return p2p_service.SendFail(incoming_packet, response=serialization.DictToBytes(self.my_info))
        if my_position_in_transaction != my_known_position:
            lg.warn('another supplier is trying to put my IDURL on another position in the family of customer %s' % self.customer_idurl)
            return p2p_service.SendFail(incoming_packet, response=serialization.DictToBytes(self.my_info))
        if _Debug:
            lg.out(_DebugLevel, '    all correct, refresh customer family for %r' % self.customer_idurl)
        self.automat('family-refresh')
        return p2p_service.SendAck(incoming_packet)

    def _on_incoming_supplier_position(self, inp):
        # this packet came from the customer, a godfather of the family ;)))
        # he told me which position in the family I should take
        incoming_packet = inp['packet']
        try:
            ecc_map = inp['customer_ecc_map']
            supplier_idurl = strng.to_bin(inp['supplier_idurl'])
            supplier_position = int(inp['supplier_position'])
            family_snapshot = id_url.to_bin_list(inp.get('family_snapshot') or [])
        except:
            lg.exc()
            return None
        if supplier_idurl != my_id.getIDURL().to_bin():
            return p2p_service.SendFail(incoming_packet, 'contacts packet with supplier position not addressed to me')
        try:
            _existing_position = self.my_info['suppliers'].index(supplier_idurl)
        except:
            _existing_position = -1
        contactsdb.add_customer_meta_info(self.customer_idurl, {
            'ecc_map': ecc_map,
            'position': supplier_position,
            'family_snapshot': id_url.to_bin_list(family_snapshot),
        })
        if _Debug:
            lg.out(_DebugLevel, 'family_member._on_incoming_supplier_position stored new meta info for customer %s:\n' % self.customer_idurl)
            lg.out(_DebugLevel, '    ecc_map=%s position=%s family_snapshot=%s' % (ecc_map, supplier_position, family_snapshot))
        return p2p_service.SendAck(incoming_packet)

    def _on_incoming_contacts_packet(self, inp):
        try:
            contacts_type = inp['type']
            incoming_packet = inp['packet']
        except:
            lg.exc()
            return None
        if contacts_type == 'suppliers_list':
            # this packet came from another supplier who belongs to the same customer family
            # he notified me about changes in the family he is trying to do
            return self._on_incoming_suppliers_list(inp)
        elif contacts_type == 'supplier_position':
            # this packet came from the customer, a godfather of the family ;)))
            # he told me which position in the family I should take
            return self._on_incoming_supplier_position(inp)
        return p2p_service.SendFail(incoming_packet, 'invalid contacts type')

    def _on_supplier_ack(self, response, info):
        if _Debug:
            lg.out(_DebugLevel, 'family_member._on_supplier_ack with %r' % response)
        if response.PacketID in self.suppliers_requests:
            self.suppliers_requests.remove(response.PacketID)
        if not self.suppliers_requests:
            self.automat('all-suppliers-agree')

    def _on_supplier_fail(self, response, info):
        if _Debug:
            lg.out(_DebugLevel, 'family_member._on_supplier_fail with %r' % response)
        if response.PacketID in self.suppliers_requests:
            self.suppliers_requests.remove(response.PacketID)
        try:
            json_payload = serialization.BytesToDict(response.Payload, keys_to_text=True)
            ecc_map = strng.to_text(json_payload['ecc_map'])
            suppliers_list = id_url.to_bin_list(json_payload['suppliers'])
        except:
            lg.exc()
            if not self.suppliers_requests:
                self.automat('all-suppliers-agree')
            return None
        if _Debug:
            lg.args(_DebugLevel, ecc_map=ecc_map, suppliers_list=suppliers_list, supplier_idurl=response.OwnerID.to_bin())
        self.automat('one-supplier-not-agree', ecc_map=ecc_map, suppliers_list=suppliers_list, supplier_idurl=response.OwnerID.to_bin())



#!/usr/bin/env python
# coins_miner.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (coins_miner.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: coins_miner.

.. role:: red

BitDust coins_miner() Automat

EVENTS:
    * :red:`accountant-connected`
    * :red:`cancel`
    * :red:`coin-confirmed`
    * :red:`coin-mined`
    * :red:`coin-rejected`
    * :red:`init`
    * :red:`lookup-failed`
    * :red:`new-data-received`
    * :red:`shutdown`
    * :red:`start`
    * :red:`stop`
    * :red:`timer-2min`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from six.moves import range

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 6

#------------------------------------------------------------------------------

import random
import string
import hashlib

from twisted.internet import reactor  # @UnresolvedImport
from twisted.internet import threads
from twisted.internet.defer import Deferred

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import sys
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.userid import my_id

from bitdust.lib import utime

from bitdust.p2p import commands
from bitdust.p2p import p2p_service

from bitdust.transport import callback

from bitdust.coins import coins_io

#------------------------------------------------------------------------------

_CoinsMiner = None

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _CoinsMiner
    if event is None and not args:
        return _CoinsMiner
    if _CoinsMiner is None:
        # set automat name and starting state here
        _CoinsMiner = CoinsMiner('coins_miner', 'AT_STARTUP', _DebugLevel, _Debug)
    if event is not None:
        _CoinsMiner.automat(event, *args, **kwargs)
    return _CoinsMiner


#------------------------------------------------------------------------------


class CoinsMiner(automat.Automat):
    """
    This class implements all the functionality of the ``coins_miner()`` state
    machine.
    """

    fast = True

    timers = {
        'timer-2min': (120, ['ACCOUNTANTS?']),
    }

    def init(self):
        """
        Method to initialize additional variables and flags at creation phase
        of coins_miner() machine.
        """
        self.offline_mode = False  # only for Debug purposes
        self.new_coin_filter_method = None
        self.connected_accountants = []
        self.min_accountants_connected = 3  # TODO: read from settings
        self.max_accountants_connected = 5  # TODO: read from settings
        self.input_data = []
        self.max_mining_counts = 10**8  # TODO: read from settings
        self.max_mining_seconds = 60*3  # TODO: read from settings
        self.simplification = 2
        self.starter_length = 10
        self.starter_limit = 9999
        self.mining_started = -1
        self.mining_counts = 0

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python
        <https://bitdust.io/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'STOPPED'
                self.doInit(*args, **kwargs)
        #---READY---
        elif self.state == 'READY':
            if event == 'stop':
                self.state = 'STOPPED'
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'new-data-received' and self.isDecideOK(*args, **kwargs):
                self.state = 'MINING'
                self.doStartMining(*args, **kwargs)
            elif event == 'new-data-received' and not self.isDecideOK(*args, **kwargs):
                self.doSendFail(*args, **kwargs)
        #---MINING---
        elif self.state == 'MINING':
            if event == 'stop':
                self.state = 'STOPPED'
                self.doStopMining(*args, **kwargs)
            elif event == 'coin-mined':
                self.state = 'PUBLISH_COIN'
                self.doSendCoinToAccountants(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doStopMining(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'new-data-received':
                self.doPushInputData(*args, **kwargs)
            elif event == 'cancel':
                self.state = 'READY'
                self.doStopMining(*args, **kwargs)
                self.doSendFail(*args, **kwargs)
                self.doPullInputData(*args, **kwargs)
        #---STOPPED---
        elif self.state == 'STOPPED':
            if event == 'start':
                self.state = 'ACCOUNTANTS?'
                self.doLookupAccountants(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        #---PUBLISH_COIN---
        elif self.state == 'PUBLISH_COIN':
            if event == 'stop':
                self.state = 'STOPPED'
            elif event == 'coin-confirmed' and self.isAllConfirmed(*args, **kwargs):
                self.state = 'READY'
                self.doSendAck(*args, **kwargs)
                self.doPullInputData(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'coin-rejected' and self.isDecideOK(*args, **kwargs):
                self.state = 'MINING'
                self.doContinueMining(*args, **kwargs)
            elif event == 'new-data-received':
                self.doPushInputData(*args, **kwargs)
            elif event == 'cancel' or (event == 'coin-rejected' and not self.isDecideOK(*args, **kwargs)):
                self.state = 'READY'
                self.doSendFail(*args, **kwargs)
                self.doPullInputData(*args, **kwargs)
        #---ACCOUNTANTS?---
        elif self.state == 'ACCOUNTANTS?':
            if event == 'accountant-connected' and not self.isMoreNeeded(*args, **kwargs):
                self.state = 'READY'
                self.doAddAccountant(*args, **kwargs)
                self.doPullInputData(*args, **kwargs)
            elif event == 'accountant-connected' and self.isMoreNeeded(*args, **kwargs):
                self.doAddAccountant(*args, **kwargs)
                self.doLookupAccountants(*args, **kwargs)
            elif event == 'new-data-received':
                self.doPushInputData(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'stop' or event == 'cancel' or event == 'timer-2min' or (event == 'lookup-failed' and not self.isAnyAccountants(*args, **kwargs)):
                self.state = 'STOPPED'
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        return None

    def isAnyAccountants(self, *args, **kwargs):
        """
        Condition method.
        """
        if self.offline_mode:
            return True
        return len(self.connected_accountants) > 0

    def isMoreNeeded(self, *args, **kwargs):
        """
        Condition method.
        """
        if self.offline_mode:
            return False
        return len(self.connected_accountants) < self.min_accountants_connected

    def isDecideOK(self, *args, **kwargs):
        """
        Condition method.
        """
        # TODO:
        return True

    def isAllConfirmed(self, *args, **kwargs):
        """
        Condition method.
        """
        if self.offline_mode:
            return True
        # TODO:
        return False

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        callback.append_inbox_callback(self._on_inbox_packet)
        if args and args[0]:
            self.new_coin_filter_method, self.offline_mode = args[0]

    def doAddAccountant(self, *args, **kwargs):
        """
        Action method.
        """
        if args and args[0]:
            if args[0] not in self.connected_accountants:
                self.connected_accountants.append(args[0])
            else:
                lg.warn('%s already connected as accountant' % args[0])

    def doLookupAccountants(self, *args, **kwargs):
        """
        Action method.
        """
        if self.offline_mode or len(self.connected_accountants) >= self.min_accountants_connected:
            self.automat('accountant-connected', '')
            return
        from bitdust.coins import accountants_finder
        accountants_finder.A('start', (self.automat, 'read'))

    def doPushInputData(self, *args, **kwargs):
        """
        Action method.
        """
        self.input_data.append(args[0])

    def doPullInputData(self, *args, **kwargs):
        """
        Action method.
        """
        if len(self.input_data) > 0:
            self.automat('new-data-received', self.input_data.pop(0))

    def doStartMining(self, *args, **kwargs):
        """
        Action method.
        """
        self.mining_started = utime.utcnow_to_sec1970()
        d = self._start(args[0])
        d.addCallback(self._on_coin_mined)
        d.addErrback(lambda err: self.automat('stop'))
        d.addErrback(lambda err: lg.exc(exc_value=err))

    def doStopMining(self, *args, **kwargs):
        """
        Action method.
        """
        self.mining_started = -1

    def doSendCoinToAccountants(self, *args, **kwargs):
        """
        Action method.
        """
        if self.offline_mode:
            self.automat('coin-confirmed')
            return
        coins = [
            args[0],
        ]
        if _Debug:
            lg.out(_DebugLevel, 'coins_miner.doSendCoinToAccountants: %s' % coins)
        for idurl in self.connected_accountants:
            p2p_service.SendCoin(idurl, coins)

    def doContinueMining(self, *args, **kwargs):
        """
        Action method.
        """

    def doSendAck(self, *args, **kwargs):
        """
        Action method.
        """
        if self.offline_mode:
            return
        # TODO:

    def doSendFail(self, *args, **kwargs):
        """
        Action method.
        """
        if self.offline_mode:
            return
        # TODO:

    def doDestroyMe(self, *args, **kwargs):
        """
        Action method.
        """
        callback.remove_inbox_callback(self._on_inbox_packet)
        self.destroy()
        global _CoinsMiner
        if self == _CoinsMiner:
            del _CoinsMiner
            _CoinsMiner = None

    #------------------------------------------------------------------------------

    def _on_inbox_packet(self, newpacket, info, status, error_message):
        if newpacket.Command == commands.Coin():
            coins_list = coins_io.read_coins_from_packet(newpacket)
            if not coins_list:
                p2p_service.SendFail(newpacket, 'failed to read coins from packet')
                return True
            if len(coins_list) != 1:
                p2p_service.SendFail(newpacket, 'expected only one coin to be mined')
                return True
            coin_json = coins_list[0]
            if not coins_io.validate_coin(coin_json):
                lg.warn('coin not valid: %s' % coin_json)
                p2p_service.SendFail(newpacket, 'coin not valid')
                return True
            if not coins_io.verify_signature(coin_json, 'creator'):
                lg.warn('creator signature is not valid: %s' % coin_json)
                p2p_service.SendFail(newpacket, 'creator signature is not valid')
                return True
            self.automat('new-data-received', coin_json)
            return True
        return False

    def _on_coin_mined(self, coin):
        if self.new_coin_filter_method is not None:
            coin = self.new_coin_filter_method(coin)
            if coin is None:
                self.automat('cancel')
                return
        self.automat('coin-mined', coin)

    def _stop_marker(self):
        if self.mining_started < 0:
            return True
        if self.mining_counts >= self.max_mining_counts:
            return True
        if utime.utcnow_to_sec1970() - self.mining_started > self.max_mining_seconds:
            return True
        self.mining_counts += 1
        return False

    def _build_starter(self, length):
        return (''.join([
            random.choice(string.uppercase + string.lowercase + string.digits)  # @UndefinedVariable
            for _ in range(length)
        ])) + '_'

    def _build_hash(self, payload):
        return hashlib.sha1(payload).hexdigest()

    def _get_hash_complexity(self, hexdigest, simplification):
        complexity = 0
        while complexity < len(hexdigest):
            if int(hexdigest[complexity], 16) < simplification:
                complexity += 1
            else:
                break
        return complexity

    def _get_hash_difficulty(self, hexdigest, simplification):
        difficulty = 0
        while True:
            ok = False
            for simpl in range(simplification):
                if hexdigest.startswith(str(simpl)*difficulty):
                    ok = True
                    break
            if ok:
                difficulty += 1
            else:
                break
        return difficulty - 1

    def _mine(self, coin_json, difficulty, simplification, starter_length, starter_limit):
        data_dump = coins_io.coin_to_string(coin_json)
        starter = self._build_starter(starter_length)
        on = 0
        while True:
            if self._stop_marker():
                if _Debug:
                    lg.out(_DebugLevel, 'coins_miner._mine STOPPED, stop marker returned True')
                return None
            check = starter + str(on)
            check += data_dump
            hexdigest = self._build_hash(check)
            if difficulty == self._get_hash_complexity(hexdigest, simplification):
                # SOLVED!
                break
            on += 1
            if on > starter_limit:
                starter = self._build_starter(starter_length)
                on = 0
        coin_json['miner'].update({
            'hash': hexdigest,
            'starter': starter + str(on),
            'mined': utime.utcnow_to_sec1970(),
        })
        return coin_json

    def _start(self, coin_json):
        coin_json['miner']['idurl'] = my_id.getIDURL().to_bin()
        # "prev" field must already be there
        prev_hash = coin_json['miner']['prev']
        difficulty = self._get_hash_difficulty(prev_hash, self.simplification)
        complexity = self._get_hash_complexity(prev_hash, self.simplification)
        if difficulty == complexity:
            complexity += 1
            if _Debug:
                lg.out(_DebugLevel, 'coins_miner.found golden coin, step up complexity: %s' % complexity)
        return threads.deferToThread(self._mine, coin_json, complexity, self.simplification, self.starter_length, self.starter_limit)


#------------------------------------------------------------------------------


def start_offline_job(coin):
    result = Deferred()
    one_miner = CoinsMiner('coins_miner', 'AT_STARTUP', _DebugLevel, _Debug)

    def _job_done(new_coin):
        reactor.callLater(0, one_miner.automat, 'shutdown')  # @UndefinedVariable
        result.callback(new_coin)
        return None

    one_miner.automat('init', (_job_done, True))
    one_miner.automat('start')
    one_miner.automat('new-data-received', coin)

    return result


#------------------------------------------------------------------------------


def _test():
    lg.set_debug_level(20)
    acoin = coins_io.storage_contract_open('http://abc.com/id.xml', 3600, 100)
    start_offline_job(acoin).addBoth(lambda *a, **kw: reactor.stop())  # @UndefinedVariable
    reactor.run()  # @UndefinedVariable


if __name__ == '__main__':
    _test()



import json
import os.path as path
from sys import exit

GENESIS_ADDRESS = 'cc846806ba14b8ec5a042d254beeeb637ee91033fa0f84c66063e0a9'


class Get:
    # "param_name":["type"] or "param_name"=["type","property_name"]
    vars = {
        'port': ['str'],
        'verify': ['bool', 'verify'],
        'testnet': ['bool'],
        'regnet': ['bool'],
        'heavy': ['bool'],
        'version': ['str', 'version'],
        'version_allow': ['list'],
        'thread_limit': ['int', 'thread_limit'],
        'rebuild_db': ['bool', 'rebuild_db'],
        'debug': ['bool', 'debug'],
        'purge': ['bool', 'purge'],
        'pause': ['int', 'pause'],
        'ledger_path': ['str', 'ledger_path'],
        'hyper_path': ['str', 'hyper_path'],
        'hyper_recompress': ['bool', 'hyper_recompress'],
        'full_ledger': ['bool', 'full_ledger'],
        'ban_threshold': ['int'],
        'tor': ['bool', 'tor'],
        'debug_level': ['str', 'debug_level'],
        'allowed': ['str', 'allowed'],
        'ram': ['bool', 'ram'],
        'node_ip': ['str', 'node_ip'],
        'light_ip': ['dict'],
        'reveal_address': ['bool'],
        'accept_peers': ['bool'],
        'banlist': ['list'],
        'whitelist': ['list'],
        'nodes_ban_reset': ['int'],
        'mempool_allowed': ['list'],
        'terminal_output': ['bool'],
        'gui_scaling': ['str'],
        'mempool_ram': ['bool'],
        'egress': ['bool'],
        'trace_db_calls': ['bool'],
        'heavy3_path': ['str'],
        'mempool_path': ['str'],
        'old_sqlite': ['bool'],
        'mandatory_message': ['list'],
    }

    # Optional default values so we don't bug if they are not in the config.
    # For compatibility
    defaults = {
        'testnet': False,
        'regnet': False,
        'heavy': True,
        'trace_db_calls': False,
        'mempool_ram': True,
        'heavy3_path': './heavy3a.bin',
        'mempool_path': './mempool.db',
        'old_sqlite': False,
        'mandatory_message': {
            'Address': 'Comment - Dict for addresses that require a message. tx to these addresses withjout a message will not be accepted by mempool.',
            'f6c0363ca1c5aa28cc584252e65a63998493ff0a5ec1bb16beda9bac': 'qTrade Exchange needs a message to route the deposit to your account',
            'd11ea307ea6de821bc28c645b1ff8dd25c6e8a9f70b3a6aeb9928754': 'VGate/ViteX Exchange needs a message to route the deposit to your account',
            '14c1b5851634f0fa8145ceea1a52cabe2443dc10350e3febf651bd3a': 'Graviex Exchange needs a message to route the deposit to your account',
            '1a174d7fdc2036e6005d93cc985424021085cc4335061307985459ce': 'Finexbox Exchange needs a message to route the deposit to your account',
            '49ca873779b36c4a503562ebf5697fca331685d79fd3deef64a46888': 'Tradesatoshi is no more listing bis but needed a message to route the deposit to your account',
            'edf2d63cdf0b6275ead22c9e6d66aa8ea31dc0ccb367fad2e7c08a25': 'Old Cryptopia address, memo',
        },  # setup here by safety, but will use the json if present for easier updates.
    }

    def load_file(self, filename):
        if not path.isfile(filename):
            filename = path.join('blockchain', 'Bismuth', filename)
        # print("Loading",filename)
        with open(filename) as fp:
            for line in fp:
                if '=' in line:
                    left, right = map(str.strip, line.rstrip('\n').split('='))
                    if 'mempool_ram_conf' == left:
                        print('Inconsistent config, param is now mempool_ram in config.txt')
                        exit()
                    if not left in self.vars:
                        # Warn for unknown param?
                        continue
                    params = self.vars[left]
                    if params[0] == 'int':
                        right = int(right)
                    elif params[0] == 'dict':
                        try:
                            right = json.loads(right)
                        except:  # compatibility
                            right = [item.strip() for item in right.split(',')]
                    elif params[0] == 'list':
                        right = [item.strip() for item in right.split(',')]
                    elif params[0] == 'bool':
                        if right.lower() in ['false', '0', '', 'no']:
                            right = False
                        else:
                            right = True

                    else:
                        # treat as "str"
                        pass
                    if len(params) > 1:
                        # deal with properties that do not match the config name.
                        left = params[1]
                    setattr(self, left, right)
        # Default genesis to keep compatibility
        self.genesis = GENESIS_ADDRESS
        for key, default in self.defaults.items():
            if key not in self.__dict__:
                setattr(self, key, default)

        # print(self.__dict__)

    def read(self, filename='config.txt', custom_filename='config_custom.txt'):
        # first of all, load from default config so we have all needed params
        self.load_file(filename)
        # then override with optional custom config
        if path.exists(custom_filename):
            self.load_file(custom_filename)
        file_name = './mandatory_message.json'
        if path.isfile(file_name):
            try:
                with open(file_name) as fp:
                    data = json.load(fp)
                    if type(data) != dict:
                        raise RuntimeWarning('Bad file format')
                    self.mandatory_message = data
                    print('mandatory_message file loaded')
            except Exception as e:
                print('Error loading mandatory_message.json {}'.format(e))
        """
        if "regnet" in self.version:
            print("Regnet, forcing ram = False")
            self.ram = False
        """



#!/usr/bin/env python
# kbucket.py
#
# Copyright (C) 2007-2008 Francois Aucamp, Meraka Institute, CSIR
# See AUTHORS for all authors and contact information.
#
# License: GNU Lesser General Public License, version 3 or later; see COPYING
#          included in this archive for details.
#
# This library is free software, distributed under the terms of
# the GNU Lesser General Public License Version 3, or any later version.
# See the COPYING file included in this archive
#
# The docstrings in this module contain epytext markup; API documentation
# may be created by processing this file with epydoc: http://epydoc.sf.net

from __future__ import absolute_import
import six

from . import constants  # @UnresolvedImport

_Debug = False


class BucketFull(Exception):
    """
    Raised when the bucket is full.
    """


class KBucket(object):
    """ Description - later
    """
    def __init__(self, rangeMin, rangeMax):
        """
        @param rangeMin: The lower boundary for the range in the 160-bit ID
                         space covered by this k-bucket
        @param rangeMax: The upper boundary for the range in the ID space
                         covered by this k-bucket
        """
        self.lastAccessed = 0
        self.rangeMin = rangeMin
        self.rangeMax = rangeMax
        self._contacts = list()

    def __repr__(self, *args, **kwargs):
        return str(self)

    def __str__(self):
        return '<KBucket %d %r to %r>' % (len(self._contacts), self.rangeMin, self.rangeMax)

    def addContact(self, contact):
        """
        Add contact to _contact list in the right order. This will move the
        contact to the end of the k-bucket if it is already present.

        @raise kademlia.kbucket.BucketFull: Raised when the bucket is full and
                                            the contact isn't in the bucket
                                            already

        @param contact: The contact to add
        @type contact: kademlia.contact.Contact
        """
        if contact in self._contacts:
            # Move the existing contact to the end of the list
            # - using the new contact to allow add-on data (e.g. optimization-specific stuff) to pe updated as well
            self._contacts.remove(contact)
            self._contacts.append(contact)
            if _Debug:
                print('[DHT KBUCKET]    moved contact to the end of the bucket %r' % contact)
        elif len(self._contacts) < constants.k:
            self._contacts.append(contact)
            if _Debug:
                print('[DHT KBUCKET]    added new contact %r' % contact)
        else:
            raise BucketFull('No space in bucket to insert contact')

    def getContact(self, contactID):
        """
        Get the contact specified node ID.
        """
        index = self._contacts.index(contactID)
        return self._contacts[index]

    def getContacts(self, count=-1, excludeContact=None):
        """
        Returns a list containing up to the first count number of contacts.

        @param count: The amount of contacts to return (if 0 or less, return
                      all contacts)
        @type count: int
        @param excludeContact: A contact to exclude; if this contact is in
                               the list of returned values, it will be
                               discarded before returning. If a C{str} is
                               passed as this argument, it must be the
                               contact's ID.
        @type excludeContact: kademlia.contact.Contact or str


        @raise IndexError: If the number of requested contacts is too large

        @return: Return up to the first count number of contacts in a list
                If no contacts are present an empty is returned
        @rtype: list
        """
        # Return all contacts in bucket
        if count <= 0:
            count = len(self._contacts)

        # Get current contact number
        currentLen = len(self._contacts)

        # If count greater than k - return only k contacts
        if count > constants.k:
            count = constants.k

        # Check if count value in range and,
        # if count number of contacts are available
        if not currentLen:
            contactList = list()

        # length of list less than requested amount
        elif currentLen < count:
            contactList = self._contacts[0:currentLen]
        # enough contacts in list
        else:
            contactList = self._contacts[0:count]

        if excludeContact in contactList:
            contactList.remove(excludeContact)

        return contactList

    def removeContact(self, contact):
        """
        Remove given contact from list.

        @param contact: The contact to remove, or a string containing the
                        contact's node ID
        @type contact: kademlia.contact.Contact or str

        @raise ValueError: The specified contact is not in this bucket
        """
        self._contacts.remove(contact)
        if _Debug:
            print('[DHT KBUCKET]    removed contact %r' % contact)

    def keyInRange(self, key):
        """
        Tests whether the specified key (i.e. node ID) is in the range of the
        160-bit ID space covered by this k-bucket (in otherwords, it returns
        whether or not the specified key should be placed in this k-bucket)

        @param key: The key to test
        @type key: str or int

        @return: C{True} if the key is in this k-bucket's range, or C{False}
                 if not.
        @rtype: bool
        """
        _k = key
        if isinstance(key, six.integer_types):
            return self.rangeMin <= key < self.rangeMax
        key = int(key, 16)
        return self.rangeMin <= key < self.rangeMax

    def __len__(self):
        return len(self._contacts)



"""
Plugin manager, based upon Michael E. Cotterell gist, MIT Licence.
Copyright 2013, Michael E. Cotterell
Copyright 2018, EggPool
Copyright 2018, BismuthFoundation

See https://github.com/bismuthfoundation/BismuthPlugins for compatible plugins and doc.
"""

import importlib
import importlib.util
import importlib.machinery
import os
import logging
import collections

__version__ = '1.0.3'


class PluginManager:
    """
    A simple plugin manager
    """
    def __init__(self, app_log=None, main_module: str = '__init__', plugin_folder: str = './plugins', config=None, verbose: bool = True, init: bool = False):
        if app_log:
            self.app_log = app_log
        else:
            logging.basicConfig(level=logging.DEBUG)
            self.app_log = logging
        self.config = config
        self.base_folder = config.ledger_path.replace('static/ledger.db', '')
        # path of mempool, colored.json and related files.
        self.plugin_folder = plugin_folder
        self.main_module = main_module
        self.verbose = verbose
        self.available_plugins = self.get_available_plugins()
        if self.verbose:
            self.app_log.info('Available plugins: {}'.format(', '.join(self.available_plugins.keys())))
        self.loaded_plugins = collections.OrderedDict({})
        if init:
            self.init()

    def init(self):
        """
        loads all available plugins and inits them.
        :return:
        """
        for plugin in self.available_plugins:
            # TODO: only load "auto load" plugins
            self.load_plugin(plugin)
        self.execute_action_hook('init', {'manager': self})

    def get_available_plugins(self):
        """
        Returns a dictionary of plugins available in the plugin folder
        """
        plugins = collections.OrderedDict({})
        try:
            for possible in os.listdir(self.plugin_folder):
                location = os.path.join(self.plugin_folder, possible)
                if os.path.isdir(location) and self.main_module + '.py' in os.listdir(location):
                    info = importlib.machinery.PathFinder().find_spec(self.main_module, [location])
                    plugins[possible] = {
                        'name': possible,
                        'info': info,
                        'autoload': True,  # Todo
                    }
        except Exception as e:
            self.app_log.info("Can't list plugins from '{}'.".format(self.plugin_folder))
        # TODO: sort by name or priority, add json specs file.
        return plugins

    def get_loaded_plugins(self):
        """
        Returns a dictionary of the loaded plugin modules
        """
        return self.loaded_plugins.copy()

    def load_plugin(self, plugin_name):
        """
        Loads a plugin module
        """
        if plugin_name in self.available_plugins:
            if plugin_name not in self.loaded_plugins:
                spec = self.available_plugins[plugin_name]['info']
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                self.loaded_plugins[plugin_name] = {'name': plugin_name, 'info': self.available_plugins[plugin_name]['info'], 'module': module}
                if self.verbose:
                    self.app_log.info("Plugin '{}' loaded".format(plugin_name))
            else:
                self.app_log.warning("Plugin '{}' already loaded".format(plugin_name))
        else:
            self.app_log.error("Cannot locate plugin '{}'".format(plugin_name))
            raise Exception("Cannot locate plugin '{}'".format(plugin_name))

    def _unload_plugin(self, plugin_name):
        del self.loaded_plugins[plugin_name]
        if self.verbose:
            self.app_log.info("Plugin '{}' unloaded".format(plugin_name))

    def unload_plugin(self, plugin_name=''):
        """
        Unloads a single plugin module or all if plugin_name is empty
        """
        try:
            if plugin_name:
                self.unload_plugin(plugin_name)
            else:
                for plugin in self.get_loaded_plugins():
                    self._unload_plugin(plugin)
        except:
            pass

    def execute_action_hook(self, hook_name, hook_params=None, first_only=False):
        """
        Executes action hook functions of the form action_hook_name contained in
        the loaded plugin modules.
        """
        for key, plugin_info in self.loaded_plugins.items():
            try:
                module = plugin_info['module']
                hook_func_name = 'action_{}'.format(hook_name)
                if hasattr(module, hook_func_name):
                    hook_func = getattr(module, hook_func_name)
                    hook_func(hook_params)
                    if first_only:
                        # Avoid deadlocks on specific use cases
                        return
            except Exception as e:
                self.app_log.warning("Plugin '{}' exception '{}' on action '{}'".format(key, e, hook_name))

    def execute_filter_hook(self, hook_name, hook_params, first_only=False):
        """
        Filters the hook_params through filter hook functions of the form
        filter_hook_name contained in the loaded plugin modules.
        """
        try:
            hook_params_keys = hook_params.keys()
            for key, plugin_info in self.loaded_plugins.items():
                try:
                    module = plugin_info['module']
                    hook_func_name = 'filter_{}'.format(hook_name)
                    if hasattr(module, hook_func_name):
                        hook_func = getattr(module, hook_func_name)
                        hook_params = hook_func(hook_params)
                        for nkey in hook_params_keys:
                            if nkey not in hook_params.keys():
                                msg = "Function '{}' in plugin '{}' is missing '{}' in the dict it returns".format(hook_func_name, plugin_info['name'], nkey)
                                self.app_log.error(msg)
                                raise Exception(msg)
                            if first_only:
                                # Avoid deadlocks on specific use cases
                                return  # will trigger the finally section
                except Exception as e:
                    self.app_log.warning("Plugin '{}' exception '{}' on filter '{}'".format(key, e, hook_name))
        except Exception as e:
            self.app_log.warning("Exception '{}' on filter '{}'".format(e, hook_name))
        finally:
            return hook_params


if __name__ == '__main__':
    print('This is Bismuth core plugin module.')
    print('See https://github.com/bismuthfoundation/BismuthPlugins for compatible plugins and doc.')



"""
Database handler module for Bismuth nodes
"""

import time
import json
import sqlite3
import essentials
import threading
from quantizer import quantize_two
import functools
from fork import Fork
import sys
import traceback


_Debug = False


def sql_trace_callback(log, id, statement):
    line = f'SQL[{id}] {statement} in {threading.current_thread()}'
    log.warning(line)


class DbHandler:
    def __init__(self, index_db, ledger_path, hyper_path, ram, ledger_ram_file, logger, trace_db_calls=False):
        self.ram = ram
        self.ledger_ram_file = ledger_ram_file
        self.hyper_path = hyper_path
        self.logger = logger
        self.trace_db_calls = trace_db_calls
        self.index_db = index_db
        self.ledger_path = ledger_path

        self.dbs = {}

        sqlite3.threadsafety = 3

        # if trace_db_calls:
        #     sqlite3.enable_callback_tracebacks(True)

        self.index = sqlite3.connect(self.index_db, timeout=1)
        if self.trace_db_calls:
            self.index.set_trace_callback(functools.partial(sql_trace_callback, self.logger.app_log, 'INDEX'))
        self.index.text_factory = str
        self.index.execute('PRAGMA case_sensitive_like = 1;')
        self.index_cursor = self.index.cursor()
        self.dbs[str(self.index)] = self.index_db

        self.hdd = sqlite3.connect(self.ledger_path, timeout=1)
        if self.trace_db_calls:
            self.hdd.set_trace_callback(functools.partial(sql_trace_callback, self.logger.app_log, 'HDD'))
        self.hdd.text_factory = str
        self.hdd.execute('PRAGMA case_sensitive_like = 1;')
        self.h = self.hdd.cursor()
        self.dbs[str(self.hdd)] = self.ledger_path

        self.hdd2 = sqlite3.connect(self.hyper_path, timeout=1)
        if self.trace_db_calls:
            self.hdd2.set_trace_callback(functools.partial(sql_trace_callback, self.logger.app_log, 'HDD2'))
        self.hdd2.text_factory = str
        self.hdd2.execute('PRAGMA case_sensitive_like = 1;')
        self.h2 = self.hdd2.cursor()
        self.dbs[str(self.hdd2)] = self.hyper_path

        if self.ram:
            self.conn = sqlite3.connect(self.ledger_ram_file, uri=True, isolation_level=None, timeout=1)
        else:
            self.conn = sqlite3.connect(self.hyper_path, uri=True, timeout=1)
        self.dbs[str(self.conn)] = 'ram'

        if self.trace_db_calls:
            self.conn.set_trace_callback(functools.partial(sql_trace_callback, self.logger.app_log, 'CONN'))
        self.conn.execute('PRAGMA journal_mode = WAL;')
        self.conn.execute('PRAGMA case_sensitive_like = 1;')
        self.conn.text_factory = str
        self.c = self.conn.cursor()

        self.SQL_TO_TRANSACTIONS = 'INSERT INTO transactions VALUES (?,?,?,?,?,?,?,?,?,?,?,?)'
        self.SQL_TO_MISC = 'INSERT INTO misc VALUES (?,?)'

        if _Debug:
            try:
                cur_connections_src = open('/tmp/db_connections', 'r').read()
            except:
                cur_connections_src = '{"l":[]}'
            cur_connections = json.loads(cur_connections_src)
            cur_connections['l'].append(threading.current_thread().name)
            open('/tmp/db_connections', 'w').write(json.dumps(cur_connections))
            self.logger.app_log.warning(f'DB_HANDLER OPEN in {threading.current_thread().name} and currently have {len(cur_connections["l"])} opened')

    def last_block_hash(self):
        self.execute(self.c, 'SELECT block_hash FROM transactions WHERE reward != 0 ORDER BY block_height DESC LIMIT 1;')
        result = self.c.fetchone()[0]
        return result

    def pubkeyget(self, address):
        self.execute_param(self.c, 'SELECT public_key FROM transactions WHERE address = ? and reward = 0 LIMIT 1', (address, ))
        result = self.c.fetchone()[0]
        return result

    def addfromalias(self, alias):
        self.execute_param(self.index_cursor, 'SELECT address FROM aliases WHERE alias = ? ORDER BY block_height ASC LIMIT 1;', (alias, ))
        try:
            address_fetch = self.index_cursor.fetchone()[0]
        except:
            address_fetch = 'No alias'
        return address_fetch

    def tokens_user(self, tokens_address):
        self.index_cursor.execute('SELECT DISTINCT token FROM tokens WHERE address OR recipient = ?', (tokens_address, ))
        result = self.index_cursor.fetchall()
        return result

    def last_block_timestamp(self):
        self.execute(self.c, 'SELECT timestamp FROM transactions WHERE reward != 0 ORDER BY block_height DESC LIMIT 1;')
        return quantize_two(self.c.fetchone()[0])

    def difflast(self):
        self.execute(self.h, 'SELECT block_height, difficulty FROM misc ORDER BY block_height DESC LIMIT 1')
        difflast = self.h.fetchone()
        return difflast

    def annverget(self, node):
        try:
            self.execute_param(self.h, 'SELECT openfield FROM transactions WHERE address = ? AND operation = ? ORDER BY block_height DESC LIMIT 1', (
                node.genesis,
                'annver',
            ))
            result = self.h.fetchone()[0]
        except:
            result = '?'
        return result

    def annget(self, node):
        try:
            self.execute_param(self.h, 'SELECT openfield FROM transactions WHERE address = ? AND operation = ? ORDER BY block_height DESC LIMIT 1', (
                node.genesis,
                'ann',
            ))
            result = self.h.fetchone()[0]
        except:
            result = 'No announcement'
        return result

    def txsearch(self, address=None, recipient=None, operation=None, openfield=None, limit=10, offset=0, block_height_from=None):
        if not address and not recipient and not operation and not openfield:
            return []
        queries = []
        params = []
        if address:
            queries.append('address = ?')
            params.append(address)
        if recipient:
            queries.append('recipient = ?')
            params.append(recipient)
        if operation:
            queries.append('operation = ?')
            params.append(operation)
        if openfield:
            queries.append('openfield = ?')
            params.append(openfield)
        if block_height_from is not None:
            queries.append('block_height >= ?')
            params.append(block_height_from)
        params.append(offset)
        params.append(limit)
        sql = 'SELECT * FROM transactions WHERE '
        sql += ' AND '.join(queries)
        sql += ' ORDER BY block_height DESC LIMIT ?, ?;'
        try:
            # print('DB:txsearch', sql, params)
            self.execute_param(self.h, sql, tuple(params))
            result = self.h.fetchall()
        except:
            traceback.print_exc()
            return []
        return result

    def block_max_ram(self):
        self.execute(self.c, 'SELECT * FROM transactions ORDER BY block_height DESC LIMIT 1')
        return essentials.format_raw_tx(self.c.fetchone())

    def aliasget(self, alias_address):
        self.execute_param(self.index_cursor, 'SELECT alias FROM aliases WHERE address = ? ', (alias_address, ))
        result = self.index_cursor.fetchall()
        if not result:
            result = [[alias_address]]
        return result

    def aliasesget(self, aliases_request):
        results = []
        for alias_address in aliases_request:
            self.execute_param(self.index_cursor, ('SELECT alias FROM aliases WHERE address = ? ORDER BY block_height ASC LIMIT 1'), (alias_address, ))
            try:
                result = self.index_cursor.fetchall()[0][0]
            except:
                result = alias_address
            results.append(result)
        return results

    def block_height_from_hash(self, data):
        try:
            self.execute_param(self.h, 'SELECT block_height FROM transactions WHERE block_hash = ?;', (data, ))
            result = self.h.fetchone()[0]
        except:
            result = None

        return result

    def blocksync(self, block):
        blocks_fetched = []
        while sys.getsizeof(str(blocks_fetched)) < 500000:  # limited size based on txs in blocks
            # db_handler.execute_param(db_handler.h, ("SELECT block_height, timestamp,address,recipient,amount,signature,public_key,keep,openfield FROM transactions WHERE block_height > ? AND block_height <= ?;"),(str(int(client_block)),) + (str(int(client_block + 1)),))
            self.execute_param(self.h, ('SELECT timestamp,address,recipient,amount,signature,public_key,operation,openfield FROM transactions WHERE block_height > ? AND block_height <= ?;'), (
                str(int(block)),
                str(int(block + 1)),
            ))
            result = self.h.fetchall()
            if not result:
                break
            blocks_fetched.extend([result])
            block = int(block) + 1
        return blocks_fetched

    def block_height_max(self):
        self.h.execute('SELECT max(block_height) FROM transactions')
        return self.h.fetchone()[0]

    def block_height_max_diff(self):
        self.h.execute('SELECT max(block_height) FROM misc')
        return self.h.fetchone()[0]

    def block_height_max_hyper(self):
        self.h2.execute('SELECT max(block_height) FROM transactions')
        return self.h2.fetchone()[0]

    def block_height_max_diff_hyper(self):
        self.h2.execute('SELECT max(block_height) FROM misc')
        return self.h2.fetchone()[0]

    def backup_higher(self, block_height):
        'backup higher blocks than given, takes data from c, which normally means RAM'
        self.execute_param(self.c, 'SELECT * FROM transactions WHERE block_height >= ?;', (block_height, ))
        backup_data = self.c.fetchall()

        self.execute_param(self.c, 'DELETE FROM transactions WHERE block_height >= ? OR block_height <= ?', (block_height, -block_height))  #this belongs to rollback_under
        self.commit(self.conn)  #this belongs to rollback_under

        self.execute_param(self.c, 'DELETE FROM misc WHERE block_height >= ?;', (block_height, ))  #this belongs to rollback_under
        self.commit(self.conn)  #this belongs to rollback_under

        return backup_data

    def rollback_under(self, block_height):
        self.h.execute('DELETE FROM transactions WHERE block_height >= ? OR block_height <= ?', (
            block_height,
            -block_height,
        ))
        self.commit(self.hdd)

        self.h.execute('DELETE FROM misc WHERE block_height >= ?', (block_height, ))
        self.commit(self.hdd)

        self.h2.execute('DELETE FROM transactions WHERE block_height >= ? OR block_height <= ?', (
            block_height,
            -block_height,
        ))
        self.commit(self.hdd2)

        self.h2.execute('DELETE FROM misc WHERE block_height >= ?', (block_height, ))
        self.commit(self.hdd2)

    def rollback_to(self, block_height):
        # We don'tt need node to have the logger
        self.logger.app_log.error('rollback_to is deprecated, use rollback_under')
        self.rollback_under(block_height)

    def tokens_rollback(self, node, height):
        """Rollback Token index

        :param height: height index of token in chain

        Simply deletes from the `tokens` table where the block_height is
        greater than or equal to the :param height: and logs the new height

        returns None
        """
        try:
            self.execute_param(self.index_cursor, 'DELETE FROM tokens WHERE block_height >= ?;', (height, ))
            self.commit(self.index)

            node.logger.app_log.warning(f'Rolled back the token index below {(height)}')
        except Exception as e:
            node.logger.app_log.warning(f'Failed to roll back the token index below {(height)} due to {e}')

    def aliases_rollback(self, node, height):
        """Rollback Alias index

        :param height: height index of token in chain

        Simply deletes from the `aliases` table where the block_height is
        greater than or equal to the :param height: and logs the new height

        returns None
        """
        try:
            self.execute_param(self.index_cursor, 'DELETE FROM aliases WHERE block_height >= ?;', (height, ))
            self.commit(self.index)

            node.logger.app_log.warning(f'Rolled back the alias index below {(height)}')
        except Exception as e:
            node.logger.app_log.warning(f'Failed to roll back the alias index below {(height)} due to {e}')

    def dev_reward(self, node, block_array, miner_tx, mining_reward, mirror_hash):
        self.execute_param(self.c, self.SQL_TO_TRANSACTIONS, (-block_array.block_height_new, str(miner_tx.q_block_timestamp), 'Development Reward', str(node.genesis), str(mining_reward), '0', '0', mirror_hash, '0', '0', '0', '0'))
        self.commit(self.conn)

    def hn_reward(self, node, block_array, miner_tx, mirror_hash):
        fork = Fork()

        if node.is_testnet and node.last_block >= fork.POW_FORK_TESTNET:
            self.reward_sum = 24 - 10*(node.last_block + 5 - fork.POW_FORK_TESTNET)/3000000

        elif node.is_mainnet and node.last_block >= fork.POW_FORK:
            self.reward_sum = 24 - 10*(node.last_block + 5 - fork.POW_FORK)/3000000
        else:
            self.reward_sum = 24

        if self.reward_sum < 0.5:
            self.reward_sum = 0.5

        self.reward_sum = '{:.8f}'.format(self.reward_sum)

        self.execute_param(
            self.c, self.SQL_TO_TRANSACTIONS,
            (-block_array.block_height_new, str(miner_tx.q_block_timestamp), 'Hypernode Payouts', '3e08b5538a4509d9daa99e01ca5912cda3e98a7f79ca01248c2bde16', self.reward_sum, '0', '0', mirror_hash, '0', '0', '0', '0')
        )
        self.commit(self.conn)

    def to_db(self, block_array, diff_save, block_transactions):
        self.execute_param(self.c, 'INSERT INTO misc VALUES (?, ?)', (block_array.block_height_new, diff_save))
        self.commit(self.conn)

        # db_handler.execute_many(db_handler.c, self.SQL_TO_TRANSACTIONS, block_transactions)

        for transaction2 in block_transactions:
            self.execute_param(
                self.c, self.SQL_TO_TRANSACTIONS, (
                    str(transaction2[0]),
                    str(transaction2[1]),
                    str(transaction2[2]),
                    str(transaction2[3]),
                    str(transaction2[4]),
                    str(transaction2[5]),
                    str(transaction2[6]),
                    str(transaction2[7]),
                    str(transaction2[8]),
                    str(transaction2[9]),
                    str(transaction2[10]),
                    str(transaction2[11]),
                )
            )
            # secure commit for slow nodes
            self.commit(self.conn)

    def db_to_drive(self, node):
        def transactions_to_h(data):
            for x in data:  # we want to save to ledger.db
                self.execute_param(self.h, self.SQL_TO_TRANSACTIONS, (x[0], x[1], x[2], x[3], x[4], x[5], x[6], x[7], x[8], x[9], x[10], x[11]))
            self.commit(self.hdd)

        def misc_to_h(data):
            for x in data:  # we want to save to ledger.db from RAM/hyper.db depending on ram conf
                self.execute_param(self.h, self.SQL_TO_MISC, (x[0], x[1]))
            self.commit(self.hdd)

        def transactions_to_h2(data):
            for x in data:
                self.execute_param(self.h2, self.SQL_TO_TRANSACTIONS, (x[0], x[1], x[2], x[3], x[4], x[5], x[6], x[7], x[8], x[9], x[10], x[11]))
            self.commit(self.hdd2)

        def misc_to_h2(data):
            for x in data:
                self.execute_param(self.h2, self.SQL_TO_MISC, (x[0], x[1]))
            self.commit(self.hdd2)

        try:
            if node.is_regnet:
                node.hdd_block = node.last_block
                node.hdd_hash = node.last_block_hash
                self.logger.app_log.warning(f'Chain: Regnet simulated move to HDD')
                return
            node.logger.app_log.warning(f'Chain: Moving new data to HDD, {node.hdd_block + 1} to {node.last_block} in {threading.current_thread()}')

            self.execute_param(
                self.c,
                'SELECT * FROM transactions '
                'WHERE block_height > ? OR block_height < ? '
                'ORDER BY block_height ASC',
                (node.hdd_block, -node.hdd_block),
            )

            result1 = self.c.fetchall()

            transactions_to_h(result1)
            if node.ram:  # we want to save to hyper.db from RAM/hyper.db depending on ram conf
                transactions_to_h2(result1)

            self.execute_param(self.c, 'SELECT * FROM misc WHERE block_height > ? ORDER BY block_height ASC', (node.hdd_block, ))
            result2 = self.c.fetchall()

            misc_to_h(result2)
            if node.ram:  # we want to save to hyper.db from RAM
                misc_to_h2(result2)

            node.hdd_block = node.last_block
            node.hdd_hash = node.last_block_hash

            node.logger.app_log.warning(f'Chain: {len(result1)} txs moved to HDD')
        except Exception as e:
            node.logger.app_log.warning(f'Chain: Exception Moving new data to HDD: {e}')
            # app_log.warning("Ledger digestion ended")  # dup with more informative digest_block notice.

    def commit(self, connection):
        """Secure commit for slow nodes"""
        sleep_delay = 0.5
        while True:
            try:
                connection.commit()
                break
            except Exception as e:
                self.logger.app_log.warning(f'Database {self.dbs.get(str(connection), "???")} connection error {e} in {threading.current_thread()}')
                self.logger.app_log.warning(f'Current threads: {",".join(map(lambda t: t.name, threading.enumerate()))}')
                time.sleep(sleep_delay)
                # sleep_delay += 1

    def execute(self, cursor, query):
        """Secure execute for slow nodes"""
        self.logger.app_log.debug(f'Execute {query} in {threading.current_thread()}')
        sleep_delay = 0.5
        while True:
            try:
                cursor.execute(query)
                break
            except sqlite3.InterfaceError as e:
                self.logger.app_log.warning(f'Database query to abort: {cursor} {query[:100]}')
                self.logger.app_log.warning(f'Database abortion reason: {e} in {threading.current_thread()}')
                break
            except sqlite3.IntegrityError as e:
                self.logger.app_log.warning(f'Database query to abort: {cursor} {query[:100]}')
                self.logger.app_log.warning(f'Database abortion reason: {e} in {threading.current_thread()}')
                break
            except Exception as e:
                self.logger.app_log.warning(f'Database query: {cursor.connection} {query[:100]}')
                self.logger.app_log.warning(f'Database retry reason: {e} in {threading.current_thread()}')
                self.logger.app_log.warning(f'Current threads: {",".join(map(lambda t: t.name, threading.enumerate()))}')
                time.sleep(sleep_delay)
                # sleep_delay += 1

    def execute_param(self, cursor, query, param):
        """Secure execute w/ param for slow nodes"""
        self.logger.app_log.debug(f'Execute with param {query} in {threading.current_thread()}')
        sleep_delay = 0.5
        while True:
            try:
                cursor.execute(query, param)
                break
            except sqlite3.InterfaceError as e:
                self.logger.app_log.warning(f'Database query to abort: {cursor} {str(query)[:100]} {str(param)[:100]}')
                self.logger.app_log.warning(f'Database abortion reason: {e} in {threading.current_thread()}')
                break
            except sqlite3.IntegrityError as e:
                self.logger.app_log.warning(f'Database query to abort: {cursor} {str(query)[:100]}')
                self.logger.app_log.warning(f'Database abortion reason: {e} in {threading.current_thread()}')
                break
            except Exception as e:
                self.logger.app_log.warning(f'Database query: {cursor.connection} {str(query)[:100]} {str(param)[:100]}')
                self.logger.app_log.warning(f'Database retry reason: {e} in {threading.current_thread()}')
                self.logger.app_log.warning(f'Current threads: {",".join(map(lambda t: t.name, threading.enumerate()))}')
                time.sleep(sleep_delay)
                # sleep_delay += 1

    def fetchall(self, cursor, query, param=None):
        """Helper to simplify calling code, execute and fetch in a single line instead of 2"""
        if param is None:
            self.execute(cursor, query)
        else:
            self.execute_param(cursor, query, param)
        return cursor.fetchall()

    def fetchone(self, cursor, query, param=None):
        """Helper to simplify calling code, execute and fetch in a single line instead of 2"""
        if param is None:
            self.execute(cursor, query)
        else:
            self.execute_param(cursor, query, param)
        res = cursor.fetchone()
        if res:
            return res[0]
        return None

    def close(self):
        try:
            self.index.close()
            self.hdd.close()
            self.hdd2.close()
            self.conn.close()
        except:
            traceback.print_exc()

        if _Debug:
            try:
                cur_connections_src = open('/tmp/db_connections', 'r').read()
            except:
                cur_connections_src = '{"l":[]}'
            cur_connections = json.loads(cur_connections_src)
            if threading.current_thread().name in cur_connections['l']:
                cur_connections['l'].remove(threading.current_thread().name)
            else:
                self.logger.app_log.warning(f'NOT FOUND opened DB connection in {threading.current_thread().name}')
            open('/tmp/db_connections', 'w').write(json.dumps(cur_connections))
            self.logger.app_log.warning(f'DB_HANDLER CLOSED in {threading.current_thread().name} and currently have {len(cur_connections["l"])} opened')



#!/usr/bin/env python
# installer.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (installer.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
#
"""
.. module:: installer.

.. role:: red

.. raw:: html

    <a href="https://bitdust.io/automats/installer/installer.png" target="_blank">
    <img src="https://bitdust.io/automats/installer/installer.png" style="max-width:100%;">
    </a>

The ``installer()`` machine is a sort of installation wizard
which is executed when user first time executes BitDust software.

It have two directions:

    1) ``register a new user`` and
    2) ``recover existing user account``

A ``id_registrator()`` and ``id_restorer()`` automats is called from here
and ``installer()`` will wait until they are finished.


EVENTS:
    * :red:`back`
    * :red:`id_registrator.state`
    * :red:`id_restorer.state`
    * :red:`init`
    * :red:`install_wizard.state`
    * :red:`load-from-file`
    * :red:`next`
    * :red:`paste-from-clipboard`
    * :red:`print`
    * :red:`recover-cmd-line`
    * :red:`recover-selected`
    * :red:`register-cmd-line`
    * :red:`register-selected`
    * :red:`register-start`
    * :red:`restore-start`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
import sys

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 4

#------------------------------------------------------------------------------

try:
    pass
except:
    sys.exit('Error initializing twisted.internet.reactor in installer.py')

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.system import bpio

from bitdust.main import settings

from bitdust.lib import misc
from bitdust.lib import nameurl

from bitdust.userid import id_registrator
from bitdust.userid import id_restorer

from bitdust.main import initializer

#------------------------------------------------------------------------------

_Installer = None

#------------------------------------------------------------------------------


def IsExist():
    global _Installer
    return _Installer is not None


#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _Installer
    if _Installer is None:
        _Installer = Installer(
            name='installer',
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_DebugLevel,
        )
    if event is not None:
        _Installer.automat(event, *args, **kwargs)
    return _Installer


#------------------------------------------------------------------------------


class Installer(automat.Automat):
    """
    A class to control the whole process of program installation.
    """

    fast = True

    output = {}
    RECOVER_RESULTS = {
        'remote_identity_not_valid': ('remote Identity is not valid', 'red'),
        'invalid_identity_source': ('incorrect source of the Identity file', 'red'),
        'invalid_identity_url': ('incorrect Identity file location', 'red'),
        'remote_identity_bad_format': ('incorrect format of the Identity file', 'red'),
        'incorrect_key': ('Private Key is not valid', 'red'),
        'idurl_not_exist': ('Identity URL address not exist or not reachable at this moment', 'blue'),
        'signing_error': ('unable to sign the local Identity file', 'red'),
        'signature_not_match': ('remote Identity and Private Key did not match', 'red'),
        'success': ('account restored!', 'green'),
    }

    def getOutput(self, state=None):
        if state is None:
            state = self.state
        return self.output.get(state, {})

    def init(self):
        self.log_events = True
        self.flagCmdLine = False

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        initializer.A('installer.state', newstate)

    def A(self, event, *args, **kwargs):
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'WHAT_TO_DO?'
                self.flagCmdLine = False
            elif event == 'register-cmd-line':
                self.state = 'REGISTER'
                self.flagCmdLine = True
                self.doInit(*args, **kwargs)
                id_registrator.A('start', *args, **kwargs)
            elif event == 'recover-cmd-line':
                self.state = 'RECOVER'
                self.flagCmdLine = True
                self.doInit(*args, **kwargs)
                id_restorer.A('start', *args, **kwargs)
        #---WHAT_TO_DO?---
        elif self.state == 'WHAT_TO_DO?':
            if event == 'register-selected':
                self.state = 'INPUT_NAME'
                self.doUpdate(*args, **kwargs)
            elif event == 'recover-selected':
                self.state = 'LOAD_KEY'
                self.doUpdate(*args, **kwargs)
        #---INPUT_NAME---
        elif self.state == 'INPUT_NAME':
            if event == 'back':
                self.state = 'WHAT_TO_DO?'
                self.doClearOutput(*args, **kwargs)
                self.doUpdate(*args, **kwargs)
            elif event == 'register-start' and self.isNameValid(*args, **kwargs):
                self.state = 'REGISTER'
                self.doClearOutput(*args, **kwargs)
                self.doSaveName(*args, **kwargs)
                id_registrator.A('start', *args, **kwargs)
                self.doUpdate(*args, **kwargs)
            elif event == 'register-start' and not self.isNameValid(*args, **kwargs):
                self.doClearOutput(*args, **kwargs)
                self.doPrintIncorrectName(*args, **kwargs)
                self.doUpdate(*args, **kwargs)
            elif event == 'print':
                self.doPrint(*args, **kwargs)
                self.doUpdate(*args, **kwargs)
        #---LOAD_KEY---
        elif self.state == 'LOAD_KEY':
            if event == 'back':
                self.state = 'WHAT_TO_DO?'
                self.doClearOutput(*args, **kwargs)
                self.doUpdate(*args, **kwargs)
            elif event == 'load-from-file':
                self.doReadKey(*args, **kwargs)
                self.doUpdate(*args, **kwargs)
            elif event == 'paste-from-clipboard':
                self.doPasteKey(*args, **kwargs)
                self.doUpdate(*args, **kwargs)
            elif event == 'restore-start':
                self.state = 'RECOVER'
                self.doClearOutput(*args, **kwargs)
                id_restorer.A('start', *args, **kwargs)
                self.doUpdate(*args, **kwargs)
            elif event == 'print':
                self.doPrint(*args, **kwargs)
                self.doUpdate(*args, **kwargs)
        #---REGISTER---
        elif self.state == 'REGISTER':
            if event == 'print':
                self.doPrint(*args, **kwargs)
                self.doUpdate(*args, **kwargs)
            elif (event == 'id_registrator.state' and args[0] == 'FAILED') and not self.flagCmdLine:
                self.state = 'INPUT_NAME'
                self.doShowOutput(*args, **kwargs)
                self.doUpdate(*args, **kwargs)
            elif (event == 'id_registrator.state' and args[0] == 'DONE') and not self.flagCmdLine:
                self.state = 'AUTHORIZED'
                self.doPrepareSettings(*args, **kwargs)
                self.doUpdate(*args, **kwargs)
            elif (event == 'id_registrator.state' and args[0] in ['DONE', 'FAILED']) and self.flagCmdLine:
                self.state = 'DONE'
                self.doUpdate(*args, **kwargs)
        #---AUTHORIZED---
        elif self.state == 'AUTHORIZED':
            if event == 'next':
                self.state = 'WIZARD'
                self.doUpdate(*args, **kwargs)
            elif event == 'print':
                self.doPrint(*args, **kwargs)
                self.doUpdate(*args, **kwargs)
        #---RECOVER---
        elif self.state == 'RECOVER':
            if event == 'print':
                self.doPrint(*args, **kwargs)
                self.doUpdate(*args, **kwargs)
            elif (event == 'id_restorer.state' and args[0] in ['RESTORED!', 'FAILED']) and self.flagCmdLine:
                self.state = 'DONE'
                self.doUpdate(*args, **kwargs)
            elif (event == 'id_restorer.state' and args[0] == 'RESTORED!') and not self.flagCmdLine:
                self.state = 'RESTORED'
                self.doRestoreSettings(*args, **kwargs)
                self.doUpdate(*args, **kwargs)
            elif (event == 'id_restorer.state' and args[0] == 'FAILED') and not self.flagCmdLine:
                self.state = 'LOAD_KEY'
                self.doUpdate(*args, **kwargs)
        #---DONE---
        elif self.state == 'DONE':
            if event == 'print':
                self.doPrint(*args, **kwargs)
                self.doUpdate(*args, **kwargs)
        #---WIZARD---
        elif self.state == 'WIZARD':
            if event == 'print':
                self.doPrint(*args, **kwargs)
                self.doUpdate(*args, **kwargs)
            elif (event == 'install_wizard.state' and args[0] == 'DONE'):
                self.state = 'DONE'
                self.doUpdate(*args, **kwargs)
        #---RESTORED---
        elif self.state == 'RESTORED':
            if event == 'print':
                self.doPrint(*args, **kwargs)
                self.doUpdate(*args, **kwargs)
            elif event == 'next':
                self.state = 'WIZARD'
                self.doUpdate(*args, **kwargs)
        return None

    def isNameValid(self, *args, **kwargs):
        if not misc.ValidUserName(args[0]['username']):
            return False
        return True

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """

    def doUpdate(self, *args, **kwargs):
        pass

    def doClearOutput(self, *args, **kwargs):
        for state in self.output.keys():
            self.output[state] = {'data': [('', 'black')]}

    def doPrint(self, *args, **kwargs):
        lg.out(8, 'installer.doPrint %s %s' % (self.state, str(*args, **kwargs)))
        if self.state not in self.output:
            self.output[self.state] = {'data': [('', 'black')]}
        if not args or args[0] is None:
            self.output[self.state]['data'] = [('', 'black')]
        else:
            self.output[self.state]['data'].append(args[0])
        if self.flagCmdLine:
            ch = '+'
            if args and args[0][1] == 'red':
                ch = '!'
            lg.out(0, '  %s %s' % (ch, args[0][0] if args else ''))

    def doShowOutput(self, *args, **kwargs):
        """
        Action method.
        """
        if 'INPUT_NAME' not in self.output:
            self.output['INPUT_NAME'] = {'data': [('', 'black')]}
        self.output['INPUT_NAME']['data'] = self.output['REGISTER']['data']

    def doPrintIncorrectName(self, *args, **kwargs):
        text, color = ('incorrect user name', 'red')
        if self.state not in self.output:
            self.output[self.state] = {'data': [('', 'black')]}
        self.output[self.state]['data'].append((text, color))
        # lg.out(0, '  [%s]' % text)

    def doSaveName(self, *args, **kwargs):
        settings.setPrivateKeySize(args[0]['pksize'])
        bpio.WriteTextFile(settings.UserNameFilename(), args[0]['username'])

    def doReadKey(self, *args, **kwargs):
        # keyfn = arg['keyfilename']
        src = args[0]['keysrc']
        lg.out(2, 'installer.doReadKey length=%s' % len(src))
        # src = bpio.ReadBinaryFile(keyfn)
        if len(src) > 1024*10:
            self.doPrint(('file is too big for private key', 'red'))
            return
        try:
            lines = src.splitlines()
            idurl = lines[0].strip()
            keysrc = '\n'.join(lines[1:])
            if idurl != nameurl.FilenameUrl(nameurl.UrlFilename(idurl)):
                idurl = ''
                keysrc = src
        except:
            lg.exc()
            idurl = ''
            keysrc = src
        if self.state not in self.output:
            self.output[self.state] = {'data': [('', 'black')]}
        self.output[self.state] = {'data': [('', 'black')]}
        self.output[self.state]['idurl'] = idurl
        self.output[self.state]['keysrc'] = keysrc
        if 'RECOVER' not in self.output:
            self.output['RECOVER'] = {'data': [('', 'black')]}
        if keysrc and idurl:
            self.output['RECOVER']['data'].append(('private key and IDURL was loaded', 'green'))
        elif not idurl and keysrc:
            self.output['RECOVER']['data'].append(('private key was loaded, provide correct IDURL now', 'blue'))

    def doPasteKey(self, *args, **kwargs):
        src = misc.getClipboardText()
        try:
            lines = src.split('\n')
            idurl = lines[0]
            keysrc = '\n'.join(lines[1:])
            if idurl != nameurl.FilenameUrl(nameurl.UrlFilename(idurl)):
                idurl = ''
                keysrc = src
        except:
            lg.exc()
            idurl = ''
            keysrc = src
        if self.state not in self.output:
            self.output[self.state] = {'data': [('', 'black')]}
        self.output[self.state]['idurl'] = idurl
        self.output[self.state]['keysrc'] = keysrc
        if 'RECOVER' not in self.output:
            self.output['RECOVER'] = {'data': [('', 'black')]}
        if keysrc and idurl:
            self.output['RECOVER']['data'].append(('private key and IDURL was loaded', 'green'))
        elif not idurl and keysrc:
            self.output['RECOVER']['data'].append(('private key was loaded, provide correct IDURL now', 'blue'))

    def doPrepareSettings(self, *args, **kwargs):
        """
        Action method.
        """

    def doRestoreSettings(self, *args, **kwargs):
        """
        Action method.
        """


#    def doIncreaseDebugLevel(self, *args, **kwargs):
#        """
#        """
#        if self.flagCmdLine and not lg.is_debug(10):
#            lg.set_debug_level(0)
#        else:
#            lg.set_debug_level(18)
